from __future__ import absolute_import, division, print_function

import pytest

sa = pytest.importorskip('sqlalchemy')
psycopg2 = pytest.importorskip('psycopg2')

import os
import itertools
import shutil

from datashape import dshape
import numpy as np
import pandas as pd

from odo.backends.csv import CSV
from odo.backends.sql import select_to_base
from odo import odo, into, resource, drop, discover, convert
from odo.utils import assert_allclose, tmpfile


@pytest.fixture(scope='module')
def pg_ip():
    return os.environ.get('POSTGRES_IP', 'localhost')

skip_no_rw_loc = \
    pytest.mark.skipif(pg_ip() != 'localhost',
                       reason=("Skipping tests with remote PG server (%s)" % pg_ip()))

names = ('tbl%d' % i for i in itertools.count())
new_schema = object()
data = [(1, 2), (10, 20), (100, 200)]
null_data = [(1, None), (10, 20), (100, 200)]


@pytest.fixture(scope='module')
def tmpdir():
    return os.environ.get('POSTGRES_TMP_DIR', None)


@pytest.yield_fixture
def csv(tmpdir):
    s = '\n'.join(','.join(map(str, row)) for row in data).encode('utf8')
    with tmpfile('.csv', dir=tmpdir) as fn:
        with open(fn, 'wb') as f:
            f.write(s)
        yield CSV(fn)


@pytest.yield_fixture
def encoding_csv(tmpdir):
    path = os.path.join(os.path.dirname(__file__), 'encoding.csv')
    with tmpfile('.csv', dir=tmpdir) as fn:
        with open(fn, 'wb') as f, open(path, 'r') as g:
            f.write(g.read().encode('latin1'))
        yield CSV(fn)


@pytest.yield_fixture
def complex_csv(tmpdir):
    path = os.path.join(os.path.dirname(__file__), 'dummydata.csv')
    with tmpfile('.csv', dir=tmpdir) as fn:
        shutil.copy(path, fn)
        os.chmod(fn, 0o777)
        yield CSV(fn, has_header=True)


@pytest.fixture
def url(pg_ip):
    return 'postgresql://postgres@{}/test::{}'.format(pg_ip, next(names))


def sql_fixture(dshape, schema=None):
    @pytest.yield_fixture(params=(True, False))
    def fixture(request, url):
        try:
            t = resource(
                url,
                dshape=dshape,
                schema=next(names) if schema is new_schema else schema,
            )
        except sa.exc.OperationalError as e:
            pytest.skip(str(e))
        else:
            try:
                bind = t.bind
                if not request.param:
                    t.metadata.bind = None
                yield t, bind
            finally:
                t.metadata.bind = bind
                drop(t)
                if t.schema:
                    bind.execute(sa.sql.ddl.DropSchema(t.schema))

    return fixture


sql = sql_fixture('var * {a: int32, b: ?int32}')
sql_with_schema = sql_fixture('var * {a: int32, b: ?int32}', new_schema)
sql_with_ugly_schema = sql_fixture('var * {a: int32, b: ?int32}', 'foo.b.ar')
complex_sql = sql_fixture("""
    var * {
        Name: string, RegistrationDate: date, ZipCode: int32, Consts: float64
    }
""")
sql_with_floats = sql_fixture('var * {a: float64, b: ?float64}')
sql_with_dts = sql_fixture('var * {a: datetime, b: ?datetime}')


@pytest.yield_fixture
def quoted_sql(pg_ip, csv):
    url = 'postgresql://postgres@{}/test::foo bar'.format(pg_ip)
    try:
        t = resource(url, dshape=discover(csv))
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@skip_no_rw_loc
def test_simple_into(csv, sql):
    sql, bind = sql
    into(sql, csv, dshape=discover(sql), bind=bind)
    assert into(list, sql, bind=bind) == data


@skip_no_rw_loc
def test_append(csv, sql):
    sql, bind = sql
    into(sql, csv, bind=bind)
    assert into(list, sql, bind=bind) == data

    into(sql, csv, bind=bind)
    assert into(list, sql, bind=bind) == data + data


def test_tryexcept_into(csv, sql):
    sql, bind = sql
    with pytest.raises(psycopg2.NotSupportedError):
        # uses multi-byte character
        into(sql, csv, quotechar="alpha", bind=bind)


@skip_no_rw_loc
def test_no_header_no_columns(csv, sql):
    sql, bind = sql
    into(sql, csv, bind=bind, dshape=discover(sql))
    assert into(list, sql, bind=bind) == data


@skip_no_rw_loc
def test_complex_into(complex_csv, complex_sql):
    complex_sql, bind = complex_sql
    # data from: http://dummydata.me/generate
    into(complex_sql, complex_csv, dshape=discover(complex_sql), bind=bind)
    assert_allclose(
        into(list, complex_sql, bind=bind), into(list, complex_csv)
    )


@skip_no_rw_loc
def test_sql_to_csv(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(csv, sql, bind=bind)
    with tmpfile('.csv', dir=tmpdir) as fn:
        csv = odo(sql, fn, bind=bind)
        assert odo(csv, list) == data
        assert discover(csv).measure.names == discover(sql).measure.names


@skip_no_rw_loc
def test_sql_select_to_csv(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(csv, sql, bind=bind)
    query = sa.select([sql.c.a])
    with tmpfile('.csv', dir=tmpdir) as fn:
        csv = odo(query, fn, bind=bind)
        assert odo(csv, list) == [(x,) for x, _ in data]


def test_invalid_escapechar(sql, csv):
    sql, bind = sql
    with pytest.raises(ValueError):
        odo(csv, sql, escapechar='12', bind=bind)

    with pytest.raises(ValueError):
        odo(csv, sql, escapechar='', bind=bind)


@skip_no_rw_loc
def test_csv_output_is_not_quoted_by_default(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(csv, sql, bind=bind)
    expected = "a,b\n1,2\n10,20\n100,200\n"
    with tmpfile('.csv', dir=tmpdir) as fn:
        csv = odo(sql, fn, bind=bind)
        with open(fn, 'rt') as f:
            result = f.read()
        assert result == expected


@skip_no_rw_loc
def test_na_value(sql, csv, tmpdir):
    sql, bind = sql
    sql = odo(null_data, sql, bind=bind)
    with tmpfile('.csv', dir=tmpdir) as fn:
        csv = odo(sql, fn, na_value='NA', bind=bind)
        with open(csv.path, 'rt') as f:
            raw = f.read()
    assert raw == 'a,b\n1,NA\n10,20\n100,200\n'


def test_different_encoding(url, encoding_csv):
    try:
        sql = odo(encoding_csv, url, encoding='latin1')
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            result = odo(sql, list)
            expected = [(u'1958.001.500131-1A', 1, None, u'', 899),
                        (u'1958.001.500156-6', 1, None, u'', 899),
                        (u'1958.001.500162-1', 1, None, u'', 899),
                        (u'1958.001.500204-2', 1, None, u'', 899),
                        (u'1958.001.500204-2A', 1, None, u'', 899),
                        (u'1958.001.500204-2B', 1, None, u'', 899),
                        (u'1958.001.500223-6', 1, None, u'', 9610),
                        (u'1958.001.500233-9', 1, None, u'', 4703),
                        (u'1909.017.000018-3', 1, 30.0, u'sumaria', 899)]
            assert result == expected
        finally:
            drop(sql)


@skip_no_rw_loc
def test_schema(csv, sql_with_schema):
    sql_with_schema, bind = sql_with_schema
    assert odo(odo(csv, sql_with_schema, bind=bind), list, bind=bind) == data


@skip_no_rw_loc
def test_ugly_schema(csv, sql_with_ugly_schema):
    sql_with_ugly_schema, bind = sql_with_ugly_schema
    assert (
        odo(odo(csv, sql_with_ugly_schema, bind=bind), list, bind=bind) ==
        data
    )


def test_schema_discover(sql_with_schema):
    sql_with_schema, bind = sql_with_schema
    sql_with_schema.metadata.bind = bind
    meta = discover(sql_with_schema.metadata)
    assert meta == dshape('{%s: var * {a: int32, b: ?int32}}' %
                          sql_with_schema.name)


@skip_no_rw_loc
def test_quoted_name(quoted_sql, csv):
    s = odo(csv, quoted_sql)
    t = odo(csv, list)
    assert sorted(odo(s, list)) == sorted(t)


def test_path_of_reduction(sql):
    sql, bind = sql
    result = convert.path(sa.select([sa.func.sum(sql.c.a)]), float)
    expected = [(sa.sql.Select, float, select_to_base)]
    assert result == expected


def test_drop_reflects_database_state(url):
    data = list(zip(range(5), range(1, 6)))

    t = odo(data, url, dshape='var * {A: int64, B: int64}')
    assert t.exists()
    assert resource(url).exists()

    drop(url)
    with pytest.raises(ValueError):
        resource(url)  # Table doesn't exist and no dshape


def test_nan_stays_nan(sql_with_floats):
    sql_with_floats, bind = sql_with_floats

    df = pd.DataFrame({'a': [np.nan, 1, 2], 'b': [3, np.nan, 4]})
    odo(df, sql_with_floats, bind=bind)
    rehydrated = odo(sql_with_floats, pd.DataFrame, bind=bind)
    pd.util.testing.assert_frame_equal(rehydrated.sort_index(axis=1), df)

    nulls_query = sa.select(sql_with_floats.c).where(
        sql_with_floats.c.a.is_(None) | sql_with_floats.c.b.is_(None)
    )
    if bind is None:
        nulls = nulls_query.execute()
    else:
        nulls = bind.execute(nulls_query)
    assert not nulls.fetchall()


def test_nat_to_null(sql_with_dts):
    sql_with_dts, bind = sql_with_dts

    df = pd.DataFrame({'a': pd.to_datetime(['2014-01-01', '2014-01-02']),
                       'b': pd.to_datetime(['2014-01-01', 'nat'])})
    odo(df, sql_with_dts, bind=bind)
    rehydrated = odo(sql_with_dts, pd.DataFrame, bind=bind)
    pd.util.testing.assert_frame_equal(rehydrated.sort_index(axis=1), df)

    nulls_query = sa.select(sql_with_dts.c).where(
        sql_with_dts.c.a.is_(None) | sql_with_dts.c.b.is_(None)
    )
    if bind is None:
        nulls = nulls_query.execute()
    else:
        nulls = bind.execute(nulls_query)
    assert nulls.fetchall() == [(pd.Timestamp('2014-01-02'), None)]


def test_to_dataframe(sql):
    sql, bind = sql
    insert_query = sql.insert().values([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])
    if bind is None:
        insert_query.execute()
    else:
        bind.execute(insert_query)

    df = odo(sql, pd.DataFrame, bind=bind)
    expected = pd.DataFrame(np.array([(1, 2), (3, 4)],
                                     dtype=[('a', 'int32'), ('b', 'float64')]))
    pd.util.testing.assert_frame_equal(df, expected)



from __future__ import absolute_import, division, print_function

from glob import glob
from .chunks import Chunks
from .resource import resource
from .utils import copydoc
from toolz import memoize, first
from datashape import discover, var
import os


class _Directory(Chunks):
    """ A directory of files on disk

    For typed containers see the ``Directory`` function which generates
    parametrized Directory classes.

    >>> from odo import CSV
    >>> c = Directory(CSV)('path/to/data/')  # doctest: +SKIP

    Normal use through resource strings

    >>> r = resource('path/to/data/*.csv')  # doctest: +SKIP
    >>> r = resource('path/to/data/')  # doctest: +SKIP


    """
    def __init__(self, path, **kwargs):
        self.path = path
        self.kwargs = kwargs

    def __iter__(self):
        return (resource(os.path.join(self.path, fn), **self.kwargs)
                    for fn in sorted(os.listdir(self.path)))


@memoize
@copydoc(_Directory)
def Directory(cls):
    """ Parametrized DirectoryClass """
    return type('Directory(%s)' % cls.__name__, (_Directory,), {'container': cls})


re_path_sep = os.path.sep
if re_path_sep == '\\':
    re_path_sep = '\\\\'

@discover.register(_Directory)
def discover_Directory(c, **kwargs):
    return var * discover(first(c)).subshape[0]


@resource.register('.+' + re_path_sep + '\*\..+', priority=15)
def resource_directory(uri, **kwargs):
    path = uri.rsplit(os.path.sep, 1)[0]
    try:
        one_uri = first(glob(uri))
    except (OSError, StopIteration):
        return _Directory(path, **kwargs)
    subtype = type(resource(one_uri, **kwargs))
    return Directory(subtype)(path, **kwargs)


@resource.register('.+' + re_path_sep, priority=9)
def resource_directory_with_trailing_slash(uri, **kwargs):
    try:
        one_uri = os.listdir(uri)[0]
    except (OSError, IndexError):
        return _Directory(uri, **kwargs)
    subtype = type(resource(one_uri, **kwargs))
    return Directory(subtype)(uri, **kwargs)



from __future__ import absolute_import, division, print_function

import pytest

import sys
import os
import numpy as np
import pandas as pd
import pandas.util.testing as tm
import gzip
import datashape
from datashape import Option, string
from collections import Iterator

from odo.backends.csv import (CSV, append, convert, resource,
                              csv_to_dataframe, CSV_to_chunks_of_dataframes,
                              infer_header)
from odo.utils import tmpfile, filetext, filetexts, raises
from odo import (into, append, convert, resource, discover, dshape, Temp,
                 chunks, odo)
from odo.temp import _Temp
from odo.compatibility import unicode


def test_csv():
    with tmpfile('.csv') as fn:
        csv = CSV(
            fn, dshape='var * {name: string, amount: int}', delimiter=',')

        assert csv.dialect['delimiter'] == ','


def test_csv_append():
    with tmpfile('.csv') as fn:
        csv = CSV(fn, has_header=False)

        data = [('Alice', 100), ('Bob', 200)]
        append(csv, data)

        assert list(convert(Iterator, csv)) == data

        with open(fn) as f:
            s = f.read()
        assert 'Alice' in s
        assert '100' in s


def test_pandas_read():
    with filetext('Alice,1\nBob,2') as fn:
        ds = datashape.dshape('var * {name: string, amount: int}')
        csv = CSV(fn)
        df = csv_to_dataframe(csv, dshape=ds)
        assert isinstance(df, pd.DataFrame)
        assert convert(list, df) == [('Alice', 1), ('Bob', 2)]
        assert list(df.columns) == ['name', 'amount']


def test_pandas_read_supports_datetimes():
    with filetext('Alice,2014-01-02\nBob,2014-01-03') as fn:
        ds = datashape.dshape('var * {name: string, when: date}')
        csv = CSV(fn)
        df = csv_to_dataframe(csv, dshape=ds)
        assert isinstance(df, pd.DataFrame)
        assert list(df.columns) == ['name', 'when']
        assert df.dtypes['when'] == 'M8[ns]'


def test_pandas_read_supports_whitespace_strings():
    with filetext('a,b, \n1,2, \n2,3, \n', extension='csv') as fn:
        csv = CSV(fn)
        ds = discover(csv)
        assert ds == datashape.dshape("var * {a: int64, b: int64, '': ?string}")


def test_pandas_read_supports_missing_integers():
    with filetext('Alice,1\nBob,') as fn:
        ds = datashape.dshape('var * {name: string, val: ?int32}')
        csv = CSV(fn)
        df = csv_to_dataframe(csv, dshape=ds)
        assert isinstance(df, pd.DataFrame)
        assert list(df.columns) == ['name', 'val']
        assert df.dtypes['val'] == 'f4'


@pytest.mark.xfail(sys.platform == 'win32' and sys.version_info[0] < 3,
                   reason="Doesn't work on Windows")
def test_pandas_read_supports_gzip():
    with filetext('Alice,1\nBob,2', open=gzip.open,
                  mode='wt', extension='.csv.gz') as fn:
        ds = datashape.dshape('var * {name: string, amount: int}')
        csv = CSV(fn)
        df = csv_to_dataframe(csv, dshape=ds)
        assert isinstance(df, pd.DataFrame)
        assert convert(list, df) == [('Alice', 1), ('Bob', 2)]
        assert list(df.columns) == ['name', 'amount']


def test_pandas_read_supports_read_csv_kwargs():
    with filetext('Alice,1\nBob,2') as fn:
        ds = datashape.dshape('var * {name: string, amount: int}')
        csv = CSV(fn)
        df = csv_to_dataframe(csv, dshape=ds, usecols=['name'])
        assert isinstance(df, pd.DataFrame)
        assert convert(list, df) == [('Alice',), ('Bob',)]


def test_pandas_write():
    with tmpfile('.csv') as fn:
        ds = datashape.dshape('var * {name: string, amount: int}')
        data = [('Alice', 1), ('Bob', 2)]
        csv = CSV(fn, has_header=True)
        append(csv, data, dshape=ds)

        with open(fn) as f:
            assert 'name' in f.read()

        # Doesn't write header twice
        append(csv, data, dshape=ds)
        with open(fn) as f:
            s = f.read()
            assert s.count('name') == 1


def test_pandas_writes_header_by_default():
    with tmpfile('.csv') as fn:
        ds = datashape.dshape('var * {name: string, amount: int}')
        data = [('Alice', 1), ('Bob', 2)]
        csv = CSV(fn)
        append(csv, data, dshape=ds)

        with open(fn) as f:
            assert 'name' in f.read()


@pytest.mark.xfail(sys.version_info[0] == 3, reason="Doesn't work on Python 3")
def test_pandas_write_gzip():
    with tmpfile('.csv.gz') as fn:
        ds = datashape.dshape('var * {name: string, amount: int}')
        data = [('Alice', 1), ('Bob', 2)]
        csv = CSV(fn, has_header=True)
        append(csv, data, dshape=ds)

        f = gzip.open(fn)
        s = f.read()
        assert 'name' in s
        assert 'Alice,1' in s
        f.close()


def test_pandas_loads_in_datetimes_naively():
    with filetext('name,when\nAlice,2014-01-01\nBob,2014-02-02') as fn:
        csv = CSV(fn, has_header=True)
        ds = datashape.dshape('var * {name: ?string, when: ?datetime}')
        assert discover(csv) == ds

        df = convert(pd.DataFrame, csv)
        assert df.dtypes['when'] == 'M8[ns]'


@pytest.mark.xfail(sys.platform == 'win32' and sys.version_info[0] < 3,
                   reason="Doesn't work on Windows")
def test_pandas_discover_on_gzipped_files():
    with filetext('name,when\nAlice,2014-01-01\nBob,2014-02-02',
                  open=gzip.open, mode='wt', extension='.csv.gz') as fn:
        csv = CSV(fn, has_header=True)
        ds = datashape.dshape('var * {name: ?string, when: ?datetime}')
        assert discover(csv) == ds


def test_csv_into_list():
    with filetext('name,val\nAlice,100\nBob,200', extension='csv') as fn:
        L = into(list, fn)
        assert L == [('Alice', 100), ('Bob', 200)]


def test_discover_csv_files_without_header():
    with filetext('Alice,2014-01-01\nBob,2014-02-02') as fn:
        csv = CSV(fn, has_header=False)
        df = convert(pd.DataFrame, csv)
        assert len(df) == 2
        assert 'Alice' not in list(df.columns)


def test_discover_csv_yields_string_on_totally_empty_columns():
    expected = dshape('var * {a: int64, b: ?string, c: int64}')
    with filetext('a,b,c\n1,,3\n4,,6\n7,,9') as fn:
        csv = CSV(fn, has_header=True)
        assert discover(csv) == expected


def test_glob():
    d = {'accounts1.csv': 'name,when\nAlice,100\nBob,200',
         'accounts2.csv': 'name,when\nAlice,300\nBob,400'}
    with filetexts(d) as fns:
        r = resource('accounts*.csv', has_header=True)
        assert convert(list, r) == [('Alice', 100), ('Bob', 200),
                                    ('Alice', 300), ('Bob', 400)]

        r = resource('*.csv')
        assert isinstance(r, chunks(CSV))


def test_pandas_csv_naive_behavior_results_in_columns():
    df = pd.DataFrame([[1, 'Alice',   100],
                       [2, 'Bob',    -200],
                       [3, 'Charlie', 300],
                       [4, 'Denis',   400],
                       [5, 'Edith',  -500]], columns=['id', 'name', 'amount'])
    with tmpfile('.csv') as fn:
        into(fn, df)

        with open(fn) as f:
            assert next(f).strip() == 'id,name,amount'


def test_discover_csv_without_columns():
    with filetext('Alice,100\nBob,200', extension='csv') as fn:
        csv = CSV(fn)
        ds = discover(csv)
        assert '100' not in str(ds)


def test_header_argument_set_with_or_without_header():
    with filetext('name,val\nAlice,100\nBob,200', extension='csv') as fn:
        assert into(list, fn) == [('Alice', 100), ('Bob', 200)]

    with filetext('Alice,100\nBob,200', extension='csv') as fn:
        assert into(list, fn) == [('Alice', 100), ('Bob', 200)]


def test_first_csv_establishes_consistent_dshape():
    d = {'accounts1.csv': 'name,when\nAlice,one\nBob,two',
         'accounts2.csv': 'name,when\nAlice,300\nBob,400'}
    with filetexts(d) as fns:
        result = into(list, 'accounts*.csv')
        assert len(result) == 4
        assert all(isinstance(val, (str, unicode)) for name, val in result)


def test_discover_csv_with_spaces_in_header():
    with filetext(' name,  val\nAlice,100\nBob,200', extension='csv') as fn:
        ds = discover(CSV(fn, has_header=True))
        assert ds.measure.names == ['name', 'val']


def test_header_disagrees_with_dshape():
    ds = datashape.dshape('var * {name: string, bal: int64}')
    with filetext('name,val\nAlice,100\nBob,200', extension='csv') as fn:
        csv = CSV(fn, header=True)
        assert convert(list, csv) == [('Alice', 100), ('Bob', 200)]

        assert list(convert(pd.DataFrame, csv).columns) == ['name', 'val']
        assert list(convert(pd.DataFrame, csv, dshape=ds).columns) == [
            'name', 'bal']


def test_header_mix_str_digits():
    ds = datashape.dshape('''var * {"On- or Off- Budget": ?string,
                                    "1990": ?string}''')
    with filetext('On- or Off- Budget,1990\nOn Budget,-628\nOff budget,"5,962"\n') as fn:
        csv = CSV(fn, has_header=True)
        df = convert(pd.DataFrame, csv)
        assert discover(csv).measure == ds.measure


def test_raise_errors_quickly_on_into_chunks_dataframe():
    with filetext('name,val\nAlice,100\nBob,foo', extension='csv') as fn:
        ds = datashape.dshape('var * {name: string, val: int}')
        csv = CSV(fn, header=True)
        assert raises(Exception,
                      lambda: CSV_to_chunks_of_dataframes(csv, dshape=ds))


def test_unused_datetime_columns():
    ds = datashape.dshape('var * {val: string, when: datetime}')
    with filetext("val,when\na,2000-01-01\nb,2000-02-02") as fn:
        csv = CSV(fn, has_header=True)
        assert convert(list, csv_to_dataframe(csv, usecols=['val'],
                                              squeeze=True, dshape=ds)) == ['a', 'b']


def test_empty_dataframe():
    with filetext('name,val', extension='csv') as fn:
        csv = CSV(fn, has_header=True)
        df = convert(pd.DataFrame, csv)
        assert isinstance(df, pd.DataFrame)


def test_csv_missing_values():
    with filetext('name,val\nAlice,100\nNA,200', extension='csv') as fn:
        csv = CSV(fn)
        assert discover(csv).measure.dict['name'] == Option(string)


def test_csv_separator_header():
    with filetext('a|b|c\n1|2|3\n4|5|6', extension='csv') as fn:
        csv = CSV(fn, delimiter='|', has_header=True)
        assert convert(list, csv) == [(1, 2, 3), (4, 5, 6)]


df = pd.DataFrame([['Alice',   100],
                   ['Bob',     200],
                   ['Charlie', 300]],
                  columns=['name', 'balance'])


def test_temp_csv():
    csv = into(Temp(CSV)('_test_temp_csv.csv'), df)
    assert isinstance(csv, CSV)

    assert into(list, csv) == into(list, df)

    del csv
    import gc
    gc.collect()
    assert not os.path.exists('_test_temp_csv.csv')


def test_convert_to_csv():
    csv = into(Temp(CSV), df)
    assert isinstance(csv, CSV)

    assert into(list, csv) == into(list, df)
    assert isinstance(csv, _Temp)


def test_unicode_column_names():
    with filetext(b'f\xc3\xbc,a\n1,2\n3,4', extension='csv', mode='wb') as fn:
        df = into(pd.DataFrame, CSV(fn, has_header=True))
    expected = pd.DataFrame([(1, 2), (3, 4)],
                            columns=[b'f\xc3\xbc'.decode('utf8'), u'a'])
    tm.assert_frame_equal(df, expected)


def test_more_unicode_column_names():
    with filetext(b'foo\xc4\x87,a\n1,2\n3,4', extension='csv',
                  mode='wb') as fn:
        df = into(pd.DataFrame, CSV(fn, has_header=True))
    expected = pd.DataFrame([(1, 2), (3, 4)],
                            columns=[b'foo\xc4\x87'.decode('utf8'), u'a'])
    tm.assert_frame_equal(df, expected)


def test_infer_header():
    with filetext('name,val\nAlice,100\nNA,200', extension='csv') as fn:
        assert infer_header(CSV(fn).path, 100) == True
    with filetext('Alice,100\nNA,200', extension='csv') as fn:
        assert infer_header(CSV(fn).path, 100) == False


def test_csv_supports_sep():
    assert CSV('foo.csv', sep=';').dialect['delimiter'] == ';'


def test_csv_to_compressed_csv():
    with tmpfile('.csv') as fn:
        with open(fn, 'w') as f:
            f.write('a,1\nb,2\nc,3')
        with tmpfile('.csv.gz') as gfn:
            result = odo(fn, gfn)
            assert odo(result, list) == odo(fn, list)


def test_has_header_on_tsv():
    with tmpfile('.csv') as fn:
        with open(fn, 'wb') as f:
            f.write(b'a\tb\n1\t2\n3\t4')
        csv = CSV(fn)
        assert csv.has_header


def test_header_with_quotes():
    csv = CSV(os.path.join(os.path.dirname(__file__), 'encoding.csv'),
              encoding='latin1')
    expected = dshape("""var * {
        D_PROC: ?string,
        NUM_SEQ: int64,
        COD_TIP_RELAC: ?float64,
        COMPL: ?string,
        COD_ASSUNTO: int64
    }
    """)
    assert discover(csv) == expected


def test_encoding_is_none():
    with tmpfile('.csv') as fn:
        with open(fn, 'w') as f:
            f.write('a,1\nb,2\nc,3'.encode('utf-8').decode('utf-8'))
        assert CSV(fn, encoding=None).encoding == 'utf-8'


def test_discover_with_dotted_names():
    with tmpfile('.csv') as fn:
        with open(fn, 'w') as f:
            f.write('a.b,c.d\n1,2\n3,4')
        dshape = discover(resource(fn))
    assert dshape == datashape.dshape('var * {"a.b": int64, "c.d": int64}')
    assert dshape.measure.names == [u'a.b', u'c.d']


try:
    unichr
except NameError:
    unichr = chr


def random_multibyte_string(nrows, string_length,
                            domain=''.join(map(unichr, range(1488, 1515)))):
    """ Generate `n` strings of length `string_length` sampled from `domain`.

    Parameters
    ----------
    n : int
        Number of random strings to generate
    string_length : int
        Length of each random string
    domain : str, optional
        The set of characters to sample from. Defaults to Hebrew.
    """
    for _ in range(nrows):
        yield ''.join(np.random.choice(list(domain), size=string_length))


@pytest.yield_fixture
def multibyte_csv():
    header = random_multibyte_string(nrows=2, string_length=3)
    single_column = random_multibyte_string(nrows=10, string_length=4)
    numbers = np.random.randint(4, size=10)
    with tmpfile('.csv') as fn:
        with open(fn, 'wb') as f:
            f.write((','.join(header) + '\n').encode('utf8'))
            f.write('\n'.join(','.join(map(unicode, row))
                              for row in zip(single_column, numbers)).encode('utf8'))
        yield fn


def test_multibyte_encoding_header(multibyte_csv):
        c = CSV(multibyte_csv, encoding='utf8', sniff_nbytes=3)
        assert c.has_header is None  # not enough data to infer header


def test_multibyte_encoding_dialect(multibyte_csv):
        c = CSV(multibyte_csv, encoding='utf8', sniff_nbytes=10)
        assert c.dialect['delimiter'] == ','


@pytest.mark.parametrize('string_dshape', ['string', 'string[25]'])
def test_string_n_convert(string_dshape):
    data = [
        '2015-03-13,FOO THE BAR',
        '2014-01-29,BAZ THE QUUX'
    ]
    ds = 'var * {k: date, n: %s}' % string_dshape
    with tmpfile('.csv') as fn:
        with open(fn, 'w') as f:
            f.write('\n'.join(data))
        csv = CSV(fn, has_header=False)
        result = odo(csv, pd.DataFrame, dshape=ds)
        assert list(result.columns) == list('kn')
    raw = [tuple(x.split(',')) for x in data]
    expected = pd.DataFrame(raw, columns=list('kn'))
    expected['k'] = pd.to_datetime(expected.k)
    tm.assert_frame_equal(result, expected)


def test_globbed_csv_to_chunks_of_dataframe():
    header = 'a,b,c\n'
    d = {'a-1.csv': header + '1,2,3\n4,5,6\n',
         'a-2.csv': header + '7,8,9\n10,11,12\n'}

    with filetexts(d):
        dfs = list(odo('a-*.csv', chunks(pd.DataFrame)))

    assert len(dfs) == 2
    columns = 'a', 'b', 'c'
    tm.assert_frame_equal(dfs[0],
                          pd.DataFrame([[1, 2, 3], [4, 5, 6]], columns=columns))
    tm.assert_frame_equal(dfs[1],
                          pd.DataFrame([[7, 8, 9], [10, 11, 12]], columns=columns))



from __future__ import absolute_import, division, print_function

import pytest
import os

pywebhdfs = pytest.importorskip('pywebhdfs')
pyhive = pytest.importorskip('pyhive')

host = os.environ.get('HDFS_TEST_HOST')
pytestmark = pytest.mark.skipif(host is None,
                                reason='No HDFS_TEST_HOST envar defined')

from pywebhdfs.webhdfs import PyWebHdfsClient

import pandas as pd
import numpy as np
import uuid
from odo.backends.hdfs import discover, HDFS, CSV, SSH, dialect_of, TableProxy
from odo.backends.sql import resource
from odo.backends.ssh import sftp
from odo import into, drop, JSONLines, odo
from odo.utils import filetext, ignoring, tmpfile
import sqlalchemy as sa
from datashape import dshape
from odo.directory import Directory
from contextlib import contextmanager

hdfs = PyWebHdfsClient(host=host, port='14000', user_name='hdfs')
ds = dshape('var * {id: ?int64, name: ?string, amount: ?int64}')
engine = resource('hive://hdfs@%s:10000/default' % host)


accounts_1_csv = """
id,name,amount
1,Alice,100
2,Bob,200
3,Charlie,300
4,Dan,400
5,Edith,500""".strip()

accounts_2_csv = """
id,name,amount
6,Frank,600
7,George,700
8,Hannah,800
""".strip()

accounts_3_csv = """
id,name,amount
9,Isaac,900
10,Jane,1000
""".strip()

@contextmanager
def accounts_data():
    a = '/user/hive/test/accounts/accounts.1.csv'
    b = '/user/hive/test/accounts/accounts.2.csv'
    c = '/user/hive/test/accounts.3.csv'
    hdfs.make_dir('user/hive/test/accounts')
    hdfs.create_file(a.lstrip('/'), accounts_1_csv)
    hdfs.create_file(b.lstrip('/'), accounts_2_csv)
    hdfs.create_file(c.lstrip('/'), accounts_3_csv)

    A = HDFS(CSV)(a, hdfs=hdfs)
    B = HDFS(CSV)(b, hdfs=hdfs)
    C = HDFS(CSV)(c, hdfs=hdfs)
    directory = HDFS(Directory(CSV))('/user/hive/test/accounts/', hdfs=hdfs)

    try:
        yield (directory, (A, B, C))
    finally:
        hdfs.delete_file_dir(a)
        hdfs.delete_file_dir(b)
        hdfs.delete_file_dir(c)

@contextmanager
def accounts_ssh():
    """ Three csv files on the remote host in a directory """
    dirname = str(uuid.uuid1())
    conn = sftp(**auth)
    conn.mkdir(dirname)
    with filetext(accounts_1_csv) as fn:
        conn.put(fn, dirname + '/accounts.1.csv')
    with filetext(accounts_2_csv) as fn:
        conn.put(fn, dirname + '/accounts.2.csv')
    with filetext(accounts_3_csv) as fn:
        conn.put(fn, dirname + '/accounts.3.csv')

    filenames = [dirname + '/accounts.%d.csv' % i for i in [1, 2, 3]]
    uris = ['ssh://ubuntu@%s:%s' % (host, fn) for fn in filenames]

    try:
        yield 'ssh://ubuntu@%s:%s/*.csv' % (host, dirname),  uris
    finally:
        for fn in filenames:
            conn.remove(fn)
        conn.rmdir(dirname)


def test_discover():
    with accounts_data() as (directory, (a, b, c)):
        assert str(discover(a)).replace('?', '') == \
                'var * {id: int64, name: string, amount: int64}'

        assert str(discover(directory)).replace('?', '') == \
                'var * {id: int64, name: string, amount: int64}'


@contextmanager
def tmpfile_hdfs(ext=''):
    fn = str(uuid.uuid1())
    if ext:
        fn = fn + '.' + ext

    try:
        yield fn
    finally:
        hdfs.delete_file_dir(fn)


def test_copy_local_files_to_hdfs():
    with tmpfile_hdfs() as target:
        with filetext('name,amount\nAlice,100\nBob,200') as source:
            csv = CSV(source)
            scsv = HDFS(CSV)(target, hdfs=hdfs)
            into(scsv, csv, blocksize=10)  # 10 bytes per message

            assert discover(scsv) == discover(csv)


def test_copy_hdfs_files_locally():
    with tmpfile('csv') as target:
        with accounts_data() as (d, (a, b, c)):
            csv = into(target, a)
            with open(csv.path) as f:
                assert f.read().strip() == accounts_1_csv


def test_copy_hdfs_data_into_memory():
    with accounts_data() as (d, (a, b, c)):
        assert into(list, a)


def test_HDFS_constructor_allows_user_alternatives():
    r = HDFS(CSV)('foo.csv', username='alice', host='host')
    assert r.hdfs.user_name == 'alice'


def test_hdfs_resource():
    r = resource('hdfs://user@hostname:1234:/path/to/myfile.json')
    assert isinstance(r, HDFS(JSONLines))
    assert r.hdfs.user_name == 'user'
    assert r.hdfs.host == 'hostname'
    assert r.hdfs.port == '1234'
    assert r.path == '/path/to/myfile.json'

    assert isinstance(resource('hdfs://path/to/myfile.csv',
                                host='host', user='user', port=1234),
                      HDFS(CSV))
    assert isinstance(resource('hdfs://path/to/*.csv',
                                host='host', user='user', port=1234),
                      HDFS(Directory(CSV)))


def normalize(s):
    return ' '.join(s.split())


auth = {'hostname': host,
        'key_filename': os.path.expanduser('~/.ssh/cdh_testing.key'),
        'username': 'ubuntu'}


@contextmanager
def hive_table(host):
    name = ('temp' + str(uuid.uuid1()).replace('-', ''))[:30]
    uri = 'hive://hdfs@%s:10000/default::%s' % (host, name)

    try:
        yield uri
    finally:
        with ignoring(Exception):
            drop(uri)


def test_hdfs_directory_hive_creation():
    with accounts_data() as (hdfs_directory, (a, b, c)):
        with hive_table(host) as uri:
            t = into(uri, hdfs_directory)
            assert isinstance(t, sa.Table)
            result = into(set, t)
            assert result > 0
            assert discover(t) == ds

            t2 = into(uri, c)  # append new singleton file
            assert len(into(list, t2)) > len(result)


def test_ssh_hive_creation():
    with hive_table(host) as uri:
        with accounts_ssh() as (_, (remote, _, _)):
            t = into(uri, remote, raise_on_errors=True, **auth)
            assert isinstance(t, sa.Table)
            assert into(set, t) == into(set, remote, **auth)

            # Load again
            t2 = into(uri, remote, raise_on_errors=True, **auth)
            assert isinstance(t2, sa.Table)
            assert len(into(list, t2)) == 2 * len(into(list, remote, **auth))


def test_hive_creation_from_local_file():
    with filetext(accounts_1_csv, extension='csv') as fn:
        with hive_table(host) as uri:
            t = into(uri, fn, **auth)
            assert isinstance(t, sa.Table)
            assert into(set, t) == into(set, fn)

            t2 = into(uri, fn, **auth)
            assert isinstance(t2, sa.Table)
            assert len(into(list, t2)) == 2 * len(into(list, fn))


def test_ssh_directory_hive_creation():
    with hive_table(host) as uri:
        with accounts_ssh() as (directory, _):
            t = odo(directory, uri, **auth)
            assert isinstance(t, sa.Table)
            assert discover(t) == ds
            assert len(into(list, t)) > 0


def test_ssh_hive_creation_with_full_urls():
    with hive_table(host) as uri:
        with accounts_ssh() as (_, (remote, _, _)):
            t = into(uri, remote,
                     key_filename=os.path.expanduser('~/.ssh/cdh_testing.key'))
            assert isinstance(t, sa.Table)
            n = len(into(list, t))
            assert n > 0

            # Load it again
            into(t, remote,
                 key_filename=os.path.expanduser('~/.ssh/cdh_testing.key'))

            # Doubles length
            assert len(into(list, t)) == 2 * n


def test_hive_resource():
    db = resource('hive://hdfs@%s:10000/default' % host)
    assert isinstance(db, sa.engine.Engine)

    db = resource('hive://%s/' % host)
    assert isinstance(db, sa.engine.Engine)
    assert str(db.url) == 'hive://hdfs@%s:10000/default' % host


def test_append_object_to_HDFS_foo():
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    with tmpfile_hdfs('json') as fn:
        js = into('hdfs://%s:%s' % (host, fn), df, hdfs=hdfs)
        assert (into(np.ndarray, js) == into(np.ndarray, df)).all()


def test_dialect_of():
    with filetext(accounts_1_csv) as fn:
        d = dialect_of(CSV(fn))
        assert d['delimiter'] == ','
        assert d['has_header'] is True

    with accounts_data() as (directory, (a, b, c)):
        directory2 = HDFS(Directory(CSV))(directory.path, hdfs=directory.hdfs)
        d = dialect_of(directory2)
        assert d['has_header'] is True

        directory2 = HDFS(Directory(CSV))(directory.path, hdfs=directory.hdfs,
                                          has_header=False)
        d = dialect_of(directory2)
        assert d['has_header'] is False


def test_hive_resource_with_internal_external():
    with hive_table(host) as uri:
        r = resource(uri, external=False, stored_as='PARQUET',
                     dshape='var * {name: string, balance: int32}')
        assert isinstance(r, sa.Table)

    with hive_table(host) as uri:
        r = resource(uri, external=False, stored_as='PARQUET')
        assert not isinstance(r, sa.Table)

    with hive_table(host) as uri:
        r = resource(uri, external=True, stored_as='PARQUET')
        assert not isinstance(r, sa.Table)


def test_copy_hive_csv_table_to_parquet():
    with hive_table(host) as csv:
        with accounts_ssh() as (_, (remote, _, _)):
            c = odo(remote, csv, **auth)
            with hive_table(host) as parquet:
                p = odo(csv, parquet, stored_as='PARQUET', external=False)
                assert odo(c, list) == odo(p, list)

            with hive_table(host) as parquet:
                try:
                    fn = '/home/hdfs/%s.parquet' % str(uuid.uuid1()).replace('-', '')[:20]
                    p = odo(csv, parquet, stored_as='PARQUET',
                            external=True, path=fn)
                    assert odo(c, list) == odo(p, list)
                finally:
                    hdfs.delete_file_dir(fn)



from __future__ import absolute_import, division, print_function

import pytest

pytest.importorskip('dask')
pytest.importorskip('pandas')

import pandas as pd
import pandas.util.testing as tm
import numpy as np
import dask.dataframe as dd
from datashape import var, Record, int64, float64, Categorical
from datashape.util.testing import assert_dshape_equal

from odo import convert, discover


def test_discover():
    df = pd.DataFrame({'x': list('a'*5 + 'b'*5 + 'c'*5),
                       'y': np.arange(15, dtype=np.int64),
                       'z': list(map(float, range(15)))},
                       columns=['x', 'y', 'z'])
    df.x = df.x.astype('category')
    ddf = dd.from_pandas(df, npartitions=2)
    assert_dshape_equal(discover(ddf),
                        var * Record([('x', Categorical(['a', 'b', 'c'])),
                                            ('y', int64), ('z', float64)]))
    assert_dshape_equal(discover(ddf.x), var * Categorical(['a', 'b', 'c']))


def test_convert():
    x = pd.DataFrame(np.arange(50).reshape(10, 5),
                     columns=list('abcde'))
    d = convert(dd.DataFrame, x, npartitions=2)
    assert isinstance(d, dd.DataFrame)


def test_convert_to_pandas_dataframe():
    x = pd.DataFrame(np.arange(50).reshape(10, 5),
                     columns=list('abcde'))
    d = convert(dd.DataFrame, x, npartitions=2)
    x2 = convert(pd.DataFrame, d)
    tm.assert_frame_equal(x2, x)


def test_convert_to_pandas_series():
    x = pd.DataFrame(np.arange(50).reshape(10, 5),
                     columns=list('abcde'))
    d = convert(dd.DataFrame, x, npartitions=2)
    a = convert(pd.Series, d.a)
    tm.assert_series_equal(a, x.a)



from __future__ import absolute_import, division, print_function

import os

def test_create_pydot_image():
    try:
        from odo.dot import dot_graph
        dot_graph()
    except:
        pass
    finally:
        for ext in ('pdf', 'png'):
            try:
                os.remove('conversions.%s' % ext)
            except:
                pass



from __future__ import print_function, division, absolute_import

import os
import uuid
import codecs

from contextlib import contextmanager, closing

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen

from toolz import memoize
from toolz.curried import take, pipe, curry, map

from .. import discover
from ..resource import resource
from ..append import append
from ..convert import convert
from ..temp import Temp
from .csv import CSV
from .json import JSON, JSONLines
from .text import TextFile

from multipledispatch import MDNotImplementedError

from .text import TextFile

from ..compatibility import urlparse
from ..utils import tmpfile, ext, sample, copydoc


class _URL(object):
    """ Parent class for data accessed through ``URLs``

    Parameters
    ----------

    url : str
        full url to file
    chunk_size : int (default 1024)
        Size of chunks streamed into memory
    decode_unicode : bool (default False)
        If True, content will be decoded using the best available
        encoding based on the response.

     Examples
    --------

    >>> from odo import URL, CSV
    >>> u = URL(CSV)('http://foo.com/myfile.csv')

    Normally create through resource uris

    >>> data = resource('http://foo.com/myfile.csv')
    >>> data.url
    'http://foo.com/myfile.csv'
    >>> data.filename
    'myfile.csv'
    """
    def __init__(self, url, chunk_size=1024, decode_unicode=False, *args,
                 **kwargs):

        self.subtype.__init__(self, url, *args, **kwargs)

        self.url = url
        self.chunk_size = chunk_size
        self.decode_unicode = decode_unicode
        self.filename = os.path.basename(urlparse(url).path)


@memoize
@copydoc(_URL)
def URL(cls):
    return type('URL(%s)' % cls.__name__, (_URL, cls), {'subtype': cls})


@sample.register((URL(CSV), URL(JSONLines)))
@contextmanager
def sample_url_line_delimited(data, lines=5, encoding='utf-8', timeout=None):
    """Get a size `length` sample from an URL CSV or URL line-delimited JSON.

    Parameters
    ----------
    data : URL(CSV)
        A hosted CSV
    lines : int, optional, default ``5``
        Number of lines to read into memory
    """

    with closing(urlopen(data.url, timeout=timeout)) as r:
        raw = pipe(r, take(lines), map(bytes.strip),
                   curry(codecs.iterdecode, encoding=encoding),
                   b'\n'.decode(encoding).join)
        with tmpfile(data.filename) as fn:
            with codecs.open(fn, 'wb', encoding=encoding) as f:
                f.write(raw)
            yield fn


@discover.register((URL(CSV), URL(JSONLines)))
def discover_url_line_delimited(c, lines=5, encoding='utf-8', **kwargs):
    """Discover CSV and JSONLines files from URL."""
    with sample(c, lines=lines, encoding=encoding) as fn:
        return discover(c.subtype(fn, **kwargs), **kwargs)


types_by_extension = {'csv': CSV, 'json': JSONLines, 'txt': TextFile}


@resource.register('ftp://.+', priority=16)
@resource.register('http://.+', priority=16)
@resource.register('https://.+', priority=16)
def resource_url(uri, **kwargs):
    path = os.path.basename(urlparse(uri).path)
    try:
        subtype = types_by_extension[ext(path)]
    except KeyError:
        subtype = type(resource(path))

    return URL(subtype)(uri, **kwargs)


@append.register(TextFile, URL(TextFile))
@append.register(JSONLines, URL(JSONLines))
@append.register(JSON, URL(JSON))
@append.register(CSV, URL(CSV))
def append_urlX_to_X(target, source, **kwargs):

    with closing(urlopen(source.url, timeout=kwargs.pop('timeout', None))) as r:
        chunk_size = 16 * source.chunk_size
        with open(target.path, 'wb') as fp:
            for chunk in iter(curry(r.read, chunk_size), b''):
                fp.write(chunk)
            return target


@convert.register(Temp(TextFile), (Temp(URL(TextFile)), URL(TextFile)))
@convert.register(Temp(JSONLines), (Temp(URL(JSONLines)), URL(JSONLines)))
@convert.register(Temp(JSON), (Temp(URL(JSON)), URL(JSON)))
@convert.register(Temp(CSV), (Temp(URL(CSV)), URL(CSV)))
def url_file_to_temp_file(data, **kwargs):
    fn = '.%s' % uuid.uuid1()
    target = Temp(data.subtype)(fn, **kwargs)
    return append(target, data, **kwargs)


@convert.register(Temp(URL(TextFile)), (TextFile, Temp(TextFile)))
@convert.register(Temp(URL(JSONLines)), (JSONLines, Temp(JSONLines)))
@convert.register(Temp(URL(JSON)), (JSON, Temp(JSON)))
@convert.register(Temp(URL(CSV)), (CSV, Temp(CSV)))
def file_to_temp_url_file(data, **kwargs):
    fn = '%s' % uuid.uuid1()
    target = Temp(URL(getattr(data, 'persistent_type', type(data))))(fn, **kwargs)
    return append(target, data, **kwargs)

try:
    from .aws import S3
except ImportError:
    pass
else:
    @append.register(S3(TextFile), URL(TextFile))
    @append.register(S3(JSON), URL(JSON))
    @append.register(S3(CSV), URL(CSV))
    @append.register(S3(JSONLines), URL(JSONLines))
    def other_remote_text_to_url_text(a, b, **kwargs):
        raise MDNotImplementedError()
try:
    from .hdfs import HDFS
except ImportError:
    pass
else:
    @append.register(HDFS(JSON), URL(JSON))
    @append.register(HDFS(TextFile), URL(TextFile))
    @append.register(HDFS(JSONLines), URL(JSONLines))
    @append.register(HDFS(JSON), URL(JSON))
    @append.register(HDFS(CSV), URL(CSV))
    def append_url_to_hdfs(target, source, **kwargs):

        path = source.filename

        try:
            subtype = types_by_extension[ext(path)]
        except KeyError:
            subtype = type(resource(path))

        t_url = resource(source.url)
        t_data = convert(Temp(subtype), t_url)
        append(target, t_data, **kwargs)



from __future__ import print_function

import pytest
import sys

pytestmark = pytest.mark.skipif(sys.platform == 'win32',
                                reason='Requires Mac or Linux')

boto = pytest.importorskip('boto')

import os
import itertools
import json
from contextlib import contextmanager, closing

import datashape
from datashape import string, float64, int64
from datashape.util.testing import assert_dshape_equal
import pandas as pd
import pandas.util.testing as tm

from odo import into, resource, S3, discover, CSV, drop, append, odo
from odo.backends.aws import get_s3_connection
from odo.utils import tmpfile
from odo.compatibility import urlopen


from boto.exception import S3ResponseError

tips_uri = 's3://nyqpug/tips.csv'

df = pd.DataFrame({
    'a': list('abc'),
    'b': [1, 2, 3],
    'c': [1.0, 2.0, 3.0]
})[['a', 'b', 'c']]


js = pd.io.json.loads(pd.io.json.dumps(df, orient='records'))

is_authorized = False
tried = False

with closing(urlopen('http://httpbin.org/ip')) as url:
    public_ip = json.loads(url.read().decode())['origin']

cidrip = public_ip + '/32'


@pytest.yield_fixture
def tmpcsv():
    with tmpfile('.csv') as fn:
        with open(fn, mode='w') as f:
            df.to_csv(f, index=False)
        yield fn


@contextmanager
def s3_bucket(extension):
    with conn():
        b = 's3://%s/%s%s' % (test_bucket_name, next(_tmps), extension)
        try:
            yield b
        finally:
            drop(resource(b))


@contextmanager
def conn():
    # requires that you have a config file or envars defined for credentials
    # this code makes me hate exceptions
    try:
        conn = get_s3_connection()
    except S3ResponseError:
        pytest.skip('unable to connect to s3')
    else:
        try:
            grants = conn.get_bucket(test_bucket_name).get_acl().acl.grants
        except S3ResponseError:
            pytest.skip('no permission to read on bucket %s' %
                        test_bucket_name)
        else:
            if not any(g.permission == 'FULL_CONTROL' or
                       g.permission == 'READ' for g in grants):
                pytest.skip('no permission to read on bucket %s' %
                            test_bucket_name)
            else:
                yield conn


test_bucket_name = 'into-redshift-csvs'

_tmps = ('tmp%d' % i for i in itertools.count())


@pytest.fixture
def s3_encryption_bucket():
    test_bucket = os.getenv('ODO_S3_ENCRYPTION_BUCKET')
    if not test_bucket:
        pytest.skip('No bucket defined that requires server-side encryption')
    return test_bucket


def test_s3_encrypted_upload(s3_encryption_bucket):
    s3_connection = boto.connect_s3()

    df = tm.makeMixedDataFrame()
    with tmpfile('.csv') as fn:
        df.to_csv(fn, index=False)
        s3_uri = 's3://{bucket}/{fn}'.format(bucket=s3_encryption_bucket, fn=os.path.basename(fn))
        odo(fn, s3_uri, s3=s3_connection, encrypt_key=True)
        result = odo(s3_uri, pd.DataFrame, s3=s3_connection)

    tm.assert_frame_equal(df, result)


def test_s3_encrypted_multipart_upload(s3_encryption_bucket):
    s3_connection = boto.connect_s3()

    df = tm.makeMixedDataFrame()
    with tmpfile('.csv') as fn:
        df.to_csv(fn, index=False)
        s3_uri = 's3://{bucket}/{fn}'.format(bucket=s3_encryption_bucket, fn=os.path.basename(fn))
        odo(fn, s3_uri, s3=s3_connection, encrypt_key=True, multipart=True)
        result = odo(s3_uri, pd.DataFrame, s3=s3_connection)

    tm.assert_frame_equal(df, result)


def test_s3_resource():
    csv = resource(tips_uri)
    assert isinstance(csv, S3(CSV))


def test_s3_discover():
    csv = resource(tips_uri)
    assert isinstance(discover(csv), datashape.DataShape)


def test_s3_to_local_csv():
    with tmpfile('.csv') as fn:
        csv = into(fn, tips_uri)
        path = os.path.abspath(csv.path)
        assert os.path.exists(path)


def test_csv_to_s3_append():
    df = tm.makeMixedDataFrame()
    with tmpfile('.csv') as fn:
        with s3_bucket('.csv') as b:
            s3 = resource(b)
            df.to_csv(fn, index=False)
            append(s3, CSV(fn))
            result = into(pd.DataFrame, s3)
    tm.assert_frame_equal(df, result)


def test_csv_to_s3_into():
    df = tm.makeMixedDataFrame()
    with tmpfile('.csv') as fn:
        with s3_bucket('.csv') as b:
            df.to_csv(fn, index=False)
            s3 = into(b, CSV(fn))
            result = into(pd.DataFrame, s3)
    tm.assert_frame_equal(df, result)


def test_frame_to_s3_to_frame():
    with s3_bucket('.csv') as b:
        s3_csv = into(b, df)
        result = into(pd.DataFrame, s3_csv)
    tm.assert_frame_equal(result, df)


def test_textfile_to_s3():
    text = 'A cow jumped over the moon'
    with tmpfile('.txt') as fn:
        with s3_bucket('.txt') as b:
            with open(fn, mode='w') as f:
                f.write(os.linesep.join(text.split()))
            result = into(b, resource(fn))
    assert discover(result) == datashape.dshape('var * string')


def test_jsonlines_to_s3():
    with tmpfile('.json') as fn:
        with open(fn, mode='w') as f:
            for row in js:
                f.write(pd.io.json.dumps(row))
                f.write(os.linesep)
        with s3_bucket('.json') as b:
            result = into(b, resource(fn))
            assert discover(result) == discover(js)


def test_s3_jsonlines_discover():
    json_dshape = discover(resource('s3://nyqpug/tips.json'))
    names = list(map(str, sorted(json_dshape.measure.names)))
    assert names == ['day', 'sex', 'size', 'smoker', 'time', 'tip',
                     'total_bill']
    types = [json_dshape.measure[name] for name in names]
    assert types == [string, string, int64, string, string, float64, float64]


def test_s3_csv_discover():
    result = discover(resource('s3://nyqpug/tips.csv'))
    expected = datashape.dshape("""var * {
      total_bill: float64,
      tip: float64,
      sex: ?string,
      smoker: ?string,
      day: ?string,
      time: ?string,
      size: int64
      }""")
    assert_dshape_equal(result, expected)


def test_s3_gz_csv_discover():
    result = discover(S3(CSV)('s3://nyqpug/tips.gz'))
    expected = datashape.dshape("""var * {
      total_bill: float64,
      tip: float64,
      sex: ?string,
      smoker: ?string,
      day: ?string,
      time: ?string,
      size: int64
      }""")
    assert_dshape_equal(result, expected)


def test_s3_to_sqlite():
    with tmpfile('.db') as fn:
        tb = into('sqlite:///%s::tips' % fn, tips_uri,
                  dshape=discover(resource(tips_uri)))
        lhs = into(list, tb)
        assert lhs == into(list, tips_uri)


def test_csv_to_s3__using_multipart_upload():
    df = pd.DataFrame({'a': ["*" * 5 * 1024 ** 2]})
    with tmpfile('.csv') as fn:
        with s3_bucket('.csv') as b:
            df.to_csv(fn, index=False)
            s3 = into(b, CSV(fn), multipart=True)
            result = into(pd.DataFrame, s3)
    tm.assert_frame_equal(df, result)


@pytest.mark.parametrize(
    ['prefix', 'suffix'],
    [
        pytest.mark.xfail(('xa', ''), raises=NotImplementedError),
        ('za', '.csv')
    ]
)
def test_chunks_of_s3(prefix, suffix):
    uri = 's3://nyqpug/{}*{}'.format(prefix, suffix)
    result = resource(uri)
    assert len(result.data) == 2
    expected = odo(tips_uri, pd.DataFrame)
    tm.assert_frame_equal(odo(result, pd.DataFrame), expected)



from __future__ import absolute_import, division, print_function

from .regex import RegexDispatcher


__all__ = 'resource'


resource = RegexDispatcher('resource')


@resource.register('.*', priority=1)
def resource_all(uri, *args, **kwargs):
    """ Refer to data by strings

    Translate a string pointing to data into a Python object pointing to that
    data.

    Filenames for common formats are valid URIs

    >>> resource('myfile.csv')  # doctest: +SKIP
    <odo.CSV ...>

    Database connection strings may embed connection information

    >>> resource('postgresql://user:pass@hostname/db::tablename')  # doctest: +SKIP
    Table('tablename', MetaData(bind=Engine(postgres://...  ... )))

    When possible ``resource`` returns an object from another popular library.
    In the case above ``resource`` gives you a ``sqlalchemy.Table`` object.

    What kinds of strings does resource support?
    --------------------------------------------

    Filenames with common formats as well as collections of those files

        myfile.csv - CSV files
        myfile.txt - Text files
        myfile.json - JSON and line-delimited JSON
        myfile.*.csv - Collections of files

    Some files, like HDF5 files or sqlite files, require a second piece of
    information, like a datapath or tablename.  We use the separator ``::`` in
    these cases.

        myfile.hdf5::/data/path
        sqlite://myfile.db::tablename

    Many systems use protocols like ``sqlite://`` to specify additional
    information.  We also use these to disambiguate when one file format, like
    HDF5, might have several possible internal formats, like standard HDF5 or
    Pandas' HDFStore

        myfile.hdf5::/data/path
        hdfstore://myfile.hdf5::/data/path

    or JSON vs JSON-lines

        json://myfile.json
        jsonlines://myfile.json

    These strings are defined as regular expressions.  See ``resource.funcs``
    to see what your installation currently supports.

    >>> resource.funcs  # doctest: +SKIP
    {'.+\.csv)(\.gz|\.bz)?': <function odo.backends.csv.resource_csv>,
     '.+\.json)(\.gz|\.bz)?': <function odo.backends.json.resource_json>,
     '\w+sql\s+://.+': <function odo.backends.sql.resource_sql>,
     ...}

    Relation with ``odo``
    ----------------------

    The main ``odo`` function uses ``resource`` to resolve string URIs.

    The following call:

    >>> odo('some-sorce', target)  # doctest: +SKIP

    is shorthand for the following:

    >>> odo(resource('some-sorce'), target)  # doctest: +SKIP

    Create datasets with resource
    -----------------------------

    Resource can also create new datasets by provding a datashape

    >>> resource('myfile.hdf5::/data', dshape='1000 * 1000 * float32')  # doctest: +SKIP
    <HDF5 dataset "data": shape (1000, 1000), type "<f4">

    To learn more about datashapes see the function ``discover``

    See Also
    --------

    odo
    discover
    """
    raise NotImplementedError("Unable to parse uri to data resource: " + uri)


@resource.register('.+::.+', priority=15)
def resource_split(uri, *args, **kwargs):
    uri, other = uri.rsplit('::', 1)
    return resource(uri, other, *args, **kwargs)



from __future__ import absolute_import, division, print_function

import os
import re
import subprocess
import sys
import decimal

from operator import attrgetter
from itertools import chain
from collections import Iterator
from datetime import datetime, date, timedelta
from distutils.spawn import find_executable

import numpy as np
import pandas as pd
import sqlalchemy as sa
from sqlalchemy import inspect
from sqlalchemy.ext.compiler import compiles
from sqlalchemy import event
from sqlalchemy.schema import CreateSchema
from sqlalchemy.dialects import mssql

from multipledispatch import MDNotImplementedError

import datashape
from datashape import DataShape, Record, Option, var, dshape, Map
from datashape.predicates import isdimension, isrecord, isscalar
from datashape import discover, datetime_, date_, float64, int64, int_, string
from datashape import float32
from datashape.dispatch import dispatch

from toolz import (partition_all, keyfilter, valfilter, identity, concat,
                   curry, merge, memoize)
from toolz.curried import pluck, map

from ..compatibility import unicode, StringIO
from ..utils import (
    keywords,
    ignoring,
    iter_except,
    filter_kwargs,
    literal_compile,
)
from ..convert import convert, ooc_types
from ..append import append
from ..resource import resource
from ..chunks import Chunks
from .csv import CSV

base = int, float, datetime, date, bool, str, decimal.Decimal, timedelta


# http://docs.sqlalchemy.org/en/latest/core/types.html

types = {
    'int64': sa.BigInteger,
    'int32': sa.Integer,
    'int': sa.Integer,
    'int16': sa.SmallInteger,
    'float32': sa.REAL,
    'float64': sa.FLOAT,
    'float': sa.FLOAT,
    'real': sa.FLOAT,
    'string': sa.Text,
    'date': sa.Date,
    'time': sa.Time,
    'datetime': sa.DateTime,
    'bool': sa.Boolean,
    "timedelta[unit='D']": sa.Interval(second_precision=0, day_precision=9),
    "timedelta[unit='h']": sa.Interval(second_precision=0, day_precision=0),
    "timedelta[unit='m']": sa.Interval(second_precision=0, day_precision=0),
    "timedelta[unit='s']": sa.Interval(second_precision=0, day_precision=0),
    "timedelta[unit='ms']": sa.Interval(second_precision=3, day_precision=0),
    "timedelta[unit='us']": sa.Interval(second_precision=6, day_precision=0),
    "timedelta[unit='ns']": sa.Interval(second_precision=9, day_precision=0),
    # ??: sa.types.LargeBinary,
}

revtypes = dict(map(reversed, types.items()))

revtypes.update({
    sa.DATETIME: datetime_,
    sa.TIMESTAMP: datetime_,
    sa.FLOAT: float64,
    sa.DATE: date_,
    sa.BIGINT: int64,
    sa.INTEGER: int_,
    sa.BIGINT: int64,
    sa.types.NullType: string,
    sa.REAL: float32,
    sa.Float: float64,
    sa.Float(precision=24): float32,
    sa.Float(precision=53): float64,
    mssql.BIT: datashape.bool_,
    mssql.DATETIMEOFFSET: string,
    mssql.MONEY: float64,
    mssql.SMALLMONEY: float32,
    mssql.UNIQUEIDENTIFIER: string
})

# interval types are special cased in discover_typeengine so remove them from
# revtypes
revtypes = valfilter(lambda x: not isinstance(x, sa.Interval), revtypes)


units_of_power = {
    0: 's',
    3: 'ms',
    6: 'us',
    9: 'ns'
}

# these aren't loaded by sqlalchemy by default
sa.dialects.registry.load('oracle')
sa.dialects.registry.load('postgresql')


def getbind(t, bind):
    if bind is None:
        return t.bind

    if isinstance(bind, sa.engine.base.Engine):
        return bind
    return create_engine(bind)


def batch(sel, chunksize=10000, bind=None):
    """Execute `sel`, streaming row at a time and fetching from the database in
    batches of size `chunksize`.

    Parameters
    ----------
    sel : sa.sql.Selectable
        Selectable to execute
    chunksize : int, optional, default 10000
        Number of rows to fetch from the database
    """

    def rowterator(sel, chunksize=chunksize):
        with getbind(sel, bind).connect() as conn:
            result = conn.execute(sel)
            yield result.keys()

            for rows in iter_except(curry(result.fetchmany, size=chunksize),
                                    sa.exc.ResourceClosedError):
                if rows:
                    yield rows
                else:
                    return
    terator = rowterator(sel)
    return next(terator), concat(terator)


@discover.register(sa.dialects.postgresql.base.INTERVAL)
def discover_postgresql_interval(t):
    return discover(sa.Interval(day_precision=0, second_precision=t.precision))


@discover.register(sa.dialects.oracle.base.INTERVAL)
def discover_oracle_interval(t):
    return discover(t.adapt(sa.Interval))


@discover.register(sa.sql.type_api.TypeEngine)
def discover_typeengine(typ):
    if isinstance(typ, sa.Interval):
        if typ.second_precision is None and typ.day_precision is None:
            return datashape.TimeDelta(unit='us')
        elif typ.second_precision == 0 and typ.day_precision == 0:
            return datashape.TimeDelta(unit='s')

        if typ.second_precision in units_of_power and not typ.day_precision:
            units = units_of_power[typ.second_precision]
        elif typ.day_precision > 0:
            units = 'D'
        else:
            raise ValueError('Cannot infer INTERVAL type with parameters'
                             'second_precision=%d, day_precision=%d' %
                             (typ.second_precision, typ.day_precision))
        return datashape.TimeDelta(unit=units)
    if typ in revtypes:
        return dshape(revtypes[typ])[0]
    if type(typ) in revtypes:
        return revtypes[type(typ)]
    if isinstance(typ, (sa.NUMERIC, sa.DECIMAL)):
        return datashape.Decimal(precision=typ.precision, scale=typ.scale)
    if isinstance(typ, (sa.String, sa.Unicode)):
        return datashape.String(typ.length, 'U8')
    else:
        for k, v in revtypes.items():
            if isinstance(k, type) and (isinstance(typ, k) or
                                        hasattr(typ, 'impl') and
                                        isinstance(typ.impl, k)):
                return v
            if k == typ:
                return v
    raise NotImplementedError("No SQL-datashape match for type %s" % typ)


@discover.register(sa.ForeignKey, sa.sql.FromClause)
def discover_foreign_key_relationship(fk, parent, parent_measure=None):
    if fk.column.table is not parent:
        parent_measure = discover(fk.column.table).measure
    return {fk.parent.name: Map(discover(fk.parent.type), parent_measure)}


@discover.register(sa.sql.elements.ColumnClause)
def discover_sqlalchemy_column(c):
    meta = Option if getattr(c, 'nullable', True) else identity
    return Record([(c.name, meta(discover(c.type)))])


@discover.register(sa.sql.FromClause)
def discover_sqlalchemy_selectable(t):
    ordering = {str(c): i for i, c in enumerate(c for c in t.columns.keys())}
    record = list(_process_columns(t.columns))
    fkeys = [discover(fkey, t, parent_measure=Record(record))
             for fkey in t.foreign_keys]
    for name, column in merge(*fkeys).items():
        record[ordering[name]] = (name, column)
    return var * Record(record)


def _process_columns(columns):
    """Process the dshapes of the columns of a table.

    Parameters
    ----------
    columns : iterable[column]
        The columns to process.

    Yields
    ------
    record_entry : tuple[str, dshape]
        A record entry containing the name and type of each column.
    """
    for col in columns:
        (name, dtype), = discover(col).fields
        yield str(name), dtype


@memoize
def metadata_of_engine(engine, schema=None):
    return sa.MetaData(engine, schema=schema)


def create_engine(uri, connect_args=None, **kwargs):
    """Creates a cached sqlalchemy engine.

    This differs from ``sa.create_engine``\s api by only accepting
    ``uri`` positionally.

    If the ``uri`` is an in memory sqlite database then this will not memioize
    the engine.
    """
    return (
        _create_engine_hashable_args
        if uri == 'sqlite:///:memory:' else
        _memoized_create_engine_hashable_args
    )(uri, connect_args=frozenset((connect_args or {}).items()), **kwargs)


def _create_engine_hashable_args(uri, connect_args=None, **kwargs):
    """Unpacks non-hashable args for ``sa.create_engine`` and puts that back
    into whatever structure is expected.
    """
    return sa.create_engine(
        uri,
        connect_args=dict(connect_args or {}),
        **kwargs
    )


_memoized_create_engine_hashable_args = memoize(_create_engine_hashable_args)


@dispatch(sa.engine.base.Engine, str)
def discover(engine, tablename):
    metadata = sa.MetaData(engine)
    if tablename not in metadata.tables:
        try:
            metadata.reflect(engine,
                             views=metadata.bind.dialect.supports_views)
        except NotImplementedError:
            metadata.reflect(engine)
    table = metadata.tables[tablename]
    return discover(table)


@dispatch(sa.engine.base.Engine)
def discover(engine):
    return discover(metadata_of_engine(engine))


@dispatch(sa.MetaData)
def discover(metadata):
    try:
        metadata.reflect(views=metadata.bind.dialect.supports_views)
    except NotImplementedError:
        metadata.reflect()
    pairs = []
    for table in sorted(metadata.tables.values(), key=attrgetter('name')):
        name = table.name
        try:
            pairs.append([name, discover(table)])
        except sa.exc.CompileError as e:
            print("Can not discover type of table %s.\n" % name +
                  "SQLAlchemy provided this error message:\n\t%s" % e.message +
                  "\nSkipping.")
        except NotImplementedError as e:
            print("Blaze does not understand a SQLAlchemy type.\n"
                  "Blaze provided the following error:\n\t%s" % "\n\t".join(e.args) +
                  "\nSkipping.")
    return DataShape(Record(pairs))


@discover.register(sa.engine.RowProxy)
def discover_row_proxy(rp):
    return Record(list(zip(rp.keys(), map(discover, rp.values()))))


def validate_foreign_keys(ds, foreign_keys):
    # passed foreign_keys and column in dshape, but not a ForeignKey type
    for field in foreign_keys:
        if field not in ds.measure.names:
            raise TypeError('Requested foreign key field %r is not a field in '
                            'datashape %s' % (field, ds))
    for field, typ in ds.measure.fields:
        if field in foreign_keys and not isinstance(getattr(typ, 'ty', typ),
                                                    Map):
            raise TypeError('Foreign key %s passed in but not a Map '
                            'datashape, got %s' % (field, typ))

        if isinstance(typ, Map) and field not in foreign_keys:
            raise TypeError('Map type %s found on column %s, but %r '
                            "wasn't found in %s" %
                            (typ, field, field, foreign_keys))


def dshape_to_table(name, ds, metadata=None, foreign_keys=None,
                    primary_key=None):
    """
    Create a SQLAlchemy table from a datashape and a name

    >>> dshape_to_table('bank', '{name: string, amount: int}') # doctest: +NORMALIZE_WHITESPACE
    Table('bank', MetaData(bind=None),
          Column('name', Text(), table=<bank>, nullable=False),
          Column('amount', Integer(), table=<bank>, nullable=False),
          schema=None)
    """

    if isinstance(ds, str):
        ds = dshape(ds)
    if not isrecord(ds.measure):
        raise TypeError('dshape measure must be a record type e.g., '
                        '"{a: int64, b: int64}". Input measure is %r' %
                        ds.measure)
    if metadata is None:
        metadata = sa.MetaData()
    if foreign_keys is None:
        foreign_keys = {}

    validate_foreign_keys(ds, foreign_keys)

    cols = dshape_to_alchemy(ds, primary_key=primary_key or frozenset())
    cols.extend(sa.ForeignKeyConstraint([column_name], [referent])
                for column_name, referent in foreign_keys.items())
    t = sa.Table(name, metadata, *cols, schema=metadata.schema)
    return attach_schema(t, t.schema)


@dispatch(object, str)
def create_from_datashape(o, ds, **kwargs):
    return create_from_datashape(o, dshape(ds), **kwargs)


@dispatch(sa.engine.base.Engine, DataShape)
def create_from_datashape(engine, ds, schema=None, foreign_keys=None,
                          primary_key=None, **kwargs):
    assert isrecord(ds), 'datashape must be Record type, got %s' % ds
    metadata = metadata_of_engine(engine, schema=schema)
    for name, sub_ds in ds[0].dict.items():
        t = dshape_to_table(name, sub_ds, metadata=metadata,
                            foreign_keys=foreign_keys,
                            primary_key=primary_key)
        t.create()
    return engine


def dshape_to_alchemy(dshape, primary_key=frozenset()):
    """

    >>> dshape_to_alchemy('int')
    <class 'sqlalchemy.sql.sqltypes.Integer'>

    >>> dshape_to_alchemy('string')
    <class 'sqlalchemy.sql.sqltypes.Text'>

    >>> dshape_to_alchemy('{name: string, amount: int}')
    [Column('name', Text(), table=None, nullable=False), Column('amount', Integer(), table=None, nullable=False)]

    >>> dshape_to_alchemy('{name: ?string, amount: ?int}')
    [Column('name', Text(), table=None), Column('amount', Integer(), table=None)]
    """
    if isinstance(dshape, str):
        dshape = datashape.dshape(dshape)
    if isinstance(dshape, Map):
        return dshape_to_alchemy(dshape.key.measure, primary_key=primary_key)
    if isinstance(dshape, Option):
        return dshape_to_alchemy(dshape.ty, primary_key=primary_key)
    if str(dshape) in types:
        return types[str(dshape)]
    if isinstance(dshape, datashape.Record):
        return [sa.Column(name,
                          dshape_to_alchemy(getattr(typ, 'ty', typ),
                                            primary_key=primary_key),
                          primary_key=name in primary_key,
                          nullable=isinstance(typ[0], Option))
                for name, typ in dshape.parameters[0]]
    if isinstance(dshape, datashape.DataShape):
        if isdimension(dshape[0]):
            return dshape_to_alchemy(dshape[1], primary_key=primary_key)
        else:
            return dshape_to_alchemy(dshape[0], primary_key=primary_key)
    if isinstance(dshape, datashape.String):
        fixlen = dshape[0].fixlen
        if fixlen is None:
            return sa.TEXT
        string_types = dict(U=sa.Unicode, A=sa.String)
        assert dshape.encoding is not None
        return string_types[dshape.encoding[0]](length=fixlen)
    if isinstance(dshape, datashape.DateTime):
        return sa.DATETIME(timezone=dshape.tz is not None)
    if isinstance(dshape, datashape.Decimal):
        return sa.NUMERIC(dshape.precision, dshape.scale)
    raise NotImplementedError("No SQLAlchemy dtype match for datashape: %s"
                              % dshape)


@convert.register(Iterator, sa.Table, cost=300.0)
def sql_to_iterator(t, bind=None, **kwargs):
    _, rows = batch(sa.select([t]), bind=bind)
    return map(tuple, rows)


@convert.register(Iterator, sa.sql.Select, cost=300.0)
def select_to_iterator(sel, dshape=None, bind=None, **kwargs):
    func = pluck(0) if dshape and isscalar(dshape.measure) else map(tuple)
    _, rows = batch(sel, bind=bind)
    return func(rows)


@convert.register(base, sa.sql.Select, cost=200.0)
def select_to_base(sel, dshape=None, bind=None, **kwargs):
    with getbind(sel, bind).connect() as conn:
        return conn.execute(sel).scalar()


@append.register(sa.Table, Iterator)
def append_iterator_to_table(t, rows, dshape=None, bind=None, **kwargs):
    assert not isinstance(t, type)
    engine = getbind(t, bind)
    if not t.exists(bind=engine):
        t.create(bind=engine)
    rows = iter(rows)

    # We see if the sequence is of tuples or dicts
    # If tuples then we coerce them to dicts
    try:
        row = next(rows)
    except StopIteration:
        return t
    rows = chain([row], rows)
    if isinstance(row, (tuple, list)):
        dshape = dshape and datashape.dshape(dshape)
        if dshape and isinstance(dshape.measure, datashape.Record):
            names = dshape.measure.names
            if set(names) != set(discover(t).measure.names):
                raise ValueError("Column names of incoming data don't match "
                                 "column names of existing SQL table\n"
                                 "Names in SQL table: %s\n"
                                 "Names from incoming data: %s\n" %
                                 (discover(t).measure.names, names))
        else:
            names = discover(t).measure.names
        rows = (dict(zip(names, row)) for row in rows)

    with engine.connect() as conn:
        for chunk in partition_all(1000, rows):  # TODO: 1000 is hardcoded
            conn.execute(t.insert(), chunk)

    return t


@append.register(sa.Table, Chunks)
def append_anything_to_sql_Table(t, c, **kwargs):
    for item in c:
        append(t, item, **kwargs)
    return t


@append.register(sa.Table, object)
def append_anything_to_sql_Table(t, o, **kwargs):
    return append(t, convert(Iterator, o, **kwargs), **kwargs)


@append.register(sa.Table, sa.Table)
def append_table_to_sql_Table(t, o, **kwargs):
    s = sa.select([o])
    return append(t, s, **kwargs)


@append.register(sa.Table, sa.sql.Select)
def append_select_statement_to_sql_Table(t, o, bind=None, **kwargs):
    t_bind = getbind(t, bind)
    o_bind = getbind(o, bind)
    if t_bind != o_bind:
        return append(
            t,
            convert(Iterator, o, bind=bind, **kwargs),
            bind=bind,
            **kwargs
        )
    bind = t_bind

    assert bind.has_table(t.name, t.schema), \
        'tables must come from the same database'

    query = t.insert().from_select(o.columns.keys(), o)

    with bind.connect() as conn:
        conn.execute(query)
    return t


def should_create_schema(ddl, target, bind, tables=None, state=None,
                         checkfirst=None, **kwargs):
    return ddl.element not in inspect(target.bind).get_schema_names()


def attach_schema(obj, schema):
    if schema is not None:
        ddl = CreateSchema(schema, quote=True)
        event.listen(
            obj,
            'before_create',
            ddl.execute_if(
                callable_=should_create_schema,
                dialect='postgresql'
            )
        )
    return obj


@resource.register(r'(.*sql.*|oracle|redshift)(\+\w+)?://.+')
def resource_sql(uri, *args, **kwargs):
    engine = create_engine(
        uri,
        # roundtrip through a frozenset of tuples so we can cache the dict
        connect_args=kwargs.pop('connect_args', {}),
        **filter_kwargs(sa.create_engine, kwargs)
    )
    ds = kwargs.pop('dshape', None)
    schema = kwargs.pop('schema', None)
    foreign_keys = kwargs.pop('foreign_keys', None)
    primary_key = kwargs.pop('primary_key', None)

    # we were also given a table name
    if args and isinstance(args[0], (str, unicode)):
        table_name, args = args[0], args[1:]
        metadata = metadata_of_engine(engine, schema=schema)

        with ignoring(sa.exc.NoSuchTableError):
            return attach_schema(
                sa.Table(
                    table_name,
                    metadata,
                    autoload_with=engine,
                ),
                schema,
            )
        if ds:
            t = dshape_to_table(table_name, ds, metadata=metadata,
                                foreign_keys=foreign_keys,
                                primary_key=primary_key)
            t.create()
            return t
        else:
            raise ValueError("Table does not exist and no dshape provided")

    # We were not given a table name
    if ds:
        create_from_datashape(engine, ds, schema=schema,
                              foreign_keys=foreign_keys)
    return engine


@resource.register('impala://.+')
def resource_impala(uri, *args, **kwargs):
    try:
        import impala.sqlalchemy
    except ImportError:
        raise ImportError("Please install or update `impyla` library")
    return resource_sql(uri, *args, **kwargs)


@resource.register('monetdb://.+')
def resource_monet(uri, *args, **kwargs):
    try:
        import monetdb
    except ImportError:
        raise ImportError("Please install the `sqlalchemy_monetdb` library")
    return resource_sql(uri, *args, **kwargs)


@resource.register('hive://.+')
def resource_hive(uri, *args, **kwargs):
    try:
        import pyhive
    except ImportError:
        raise ImportError("Please install the `PyHive` library.")

    pattern = 'hive://((?P<user>[a-zA-Z_]\w*)@)?(?P<host>[\w.]+)(:(?P<port>\d*))?(/(?P<database>\w*))?'
    d = re.search(pattern, uri.split('::')[0]).groupdict()

    defaults = {'port': '10000',
                'user': 'hdfs',
                'database': 'default'}
    for k, v in d.items():
        if not v:
            d[k] = defaults[k]

    if d['user']:
        d['user'] += '@'

    uri2 = 'hive://%(user)s%(host)s:%(port)s/%(database)s' % d
    if '::' in uri:
        uri2 += '::' + uri.split('::')[1]

    return resource_sql(uri2, *args, **kwargs)


ooc_types.add(sa.Table)


@dispatch(sa.Table)
def drop(table, bind=None):
    bind = getbind(table, bind)
    table.drop(bind=bind, checkfirst=True)
    if table.exists(bind=bind):
        raise ValueError('table %r dropped but still exists' % table.name)
    metadata_of_engine(bind, schema=table.schema).remove(table)


@convert.register(sa.sql.Select, sa.Table, cost=0)
def table_to_select(t, **kwargs):
    return t.select()


@convert.register(pd.DataFrame, (sa.sql.Select, sa.sql.Selectable), cost=300.0)
def select_or_selectable_to_frame(el, bind=None, dshape=None, **kwargs):
    bind = getbind(el, bind)
    if bind.dialect.name == 'postgresql':
        buf = StringIO()
        append(CSV(None, buffer=buf), el, bind=bind, **kwargs)
        buf.seek(0)
        datetime_fields = []
        other_dtypes = {}
        try:
            fields = dshape.measure.fields
        except AttributeError:
            fields = [(0, dshape.measure)]

        for field, dtype in fields:
            if getattr(dtype, 'ty', dtype) == datashape.datetime_:
                datetime_fields.append(field)
            elif isinstance(dtype, Option):
                ty = dtype.ty
                if ty in datashape.integral:
                    other_dtypes[field] = 'float64'
                else:
                    other_dtypes[field] = ty.to_numpy_dtype()
            else:
                other_dtypes[field] = dtype.to_numpy_dtype()

        return pd.read_csv(
            buf,
            parse_dates=datetime_fields,
            dtype=other_dtypes,
            skip_blank_lines=False,
        )

    columns, rows = batch(el, bind=bind)
    dtypes = {}
    try:
        fields = dshape.measure.fields
    except AttributeError:
        fields = [(columns[0], dshape.measure)]
    for field, dtype in fields:
        if isinstance(dtype, Option):
            ty = dtype.ty
            if ty in datashape.integral:
                dtypes[field] = 'float64'
            else:
                dtypes[field] = ty.to_numpy_dtype()
        else:
            dtypes[field] = dtype.to_numpy_dtype()

    return pd.DataFrame(np.array(list(map(tuple, rows)),
                                 dtype=[(str(c), dtypes[c]) for c in columns]))


class CopyToCSV(sa.sql.expression.Executable, sa.sql.ClauseElement):

    def __init__(
        self,
        element,
        path,
        delimiter=',',
        quotechar='"',
        lineterminator='\n',
        escapechar='\\',
        header=True,
        na_value='',
        encoding=None,
        bind=None,
    ):
        self.element = element
        self.path = path
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.lineterminator = lineterminator
        self._bind = bind = getbind(element, bind)

        # mysql cannot write headers
        self.header = header and bind.dialect.name != 'mysql'
        self.escapechar = escapechar
        self.na_value = na_value
        self.encoding = encoding

    @property
    def bind(self):
        return self._bind


@compiles(CopyToCSV, 'postgresql')
def compile_copy_to_csv_postgres(element, compiler, **kwargs):
    selectable = element.element
    return compiler.process(
        sa.text(
            """COPY {0} TO STDOUT
            WITH (
                FORMAT CSV,
                HEADER :header,
                DELIMITER :delimiter,
                QUOTE :quotechar,
                NULL :na_value,
                ESCAPE :escapechar,
                ENCODING :encoding
            )
            """.format(
                compiler.preparer.format_table(selectable)
                if isinstance(selectable, sa.Table)
                else '({0})'.format(literal_compile(selectable))
            )
        ).bindparams(
            header=element.header,
            delimiter=element.delimiter,
            quotechar=element.quotechar,
            na_value=element.na_value,
            escapechar=element.escapechar,
            encoding=element.encoding or element.bind.execute(
                'show client_encoding'
            ).scalar()
        ),
        **kwargs
    )


@compiles(CopyToCSV, 'mysql')
def compile_copy_to_csv_mysql(element, compiler, **kwargs):
    selectable = element.element
    return compiler.process(
        sa.text(
            """{0} INTO OUTFILE :path
            CHARACTER SET :encoding
            FIELDS TERMINATED BY :delimiter
            OPTIONALLY ENCLOSED BY :quotechar
            ESCAPED BY :escapechar
            LINES TERMINATED BY :lineterminator
            """.format(
                compiler.process(
                    selectable.select()
                    if isinstance(selectable, sa.Table) else selectable,
                    **kwargs
                )
            )
        ).bindparams(
            path=element.path,
            encoding=element.encoding or element.bind.execute(
                'select @@character_set_client'
            ).scalar(),
            delimiter=element.delimiter,
            quotechar=element.quotechar,
            escapechar=element.escapechar,
            lineterminator=element.lineterminator
        )
    )


@compiles(CopyToCSV, 'sqlite')
def compile_copy_to_csv_sqlite(element, compiler, **kwargs):
    if element.encoding is not None:
        raise ValueError(
            "'encoding' keyword argument not supported for "
            "SQLite to CSV conversion"
        )
    if not find_executable('sqlite3'):
        raise MDNotImplementedError("Could not find sqlite executable")

    # we are sending a SQL string directorly to the SQLite process so we always
    # need to bind everything before sending it
    kwargs['literal_binds'] = True

    selectable = element.element
    sql = compiler.process(
        selectable.select() if isinstance(selectable, sa.Table) else selectable,
        **kwargs
    ) + ';'
    sql = re.sub(r'\s{2,}', ' ', re.sub(r'\s*\n\s*', ' ', sql)).encode(
        sys.getfilesystemencoding()  # we send bytes to the process
    )
    cmd = ['sqlite3', '-csv',
           '-%sheader' % ('no' if not element.header else ''),
           '-separator', element.delimiter,
           selectable.bind.url.database]
    with open(element.path, mode='at') as f:
        subprocess.Popen(cmd, stdout=f, stdin=subprocess.PIPE).communicate(sql)

    # This will be a no-op since we're doing the write during the compile
    return ''


@append.register(CSV, sa.sql.Selectable)
def append_table_to_csv(csv, selectable, dshape=None, bind=None, **kwargs):
    kwargs = keyfilter(keywords(CopyToCSV).__contains__,
                       merge(csv.dialect, kwargs))
    stmt = CopyToCSV(
        selectable,
        os.path.abspath(csv.path) if csv.path is not None else None,
        bind=bind,
        **kwargs
    )

    bind = getbind(selectable, bind)
    if bind.dialect.name == 'postgresql':
        with csv.open('ab+') as f:
            with bind.begin() as conn:
                conn.connection.cursor().copy_expert(literal_compile(stmt), f)
    else:
        with bind.begin() as conn:
            conn.execute(stmt)
    return csv


try:
    from .hdfs import HDFS
except ImportError:
    pass
else:
    @append.register(HDFS(CSV), sa.sql.Selectable)
    def append_selectable_to_hdfs_csv(*args, **kwargs):
        raise MDNotImplementedError()



from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd
from datashape.predicates import isscalar
from toolz import concat, partition_all, compose
from collections import Iterator, Iterable
import datashape
from datashape import discover
from .core import NetworkDispatcher, ooc_types
from .chunks import chunks, Chunks
from .numpy_dtype import dshape_to_numpy
from .utils import records_to_tuples
from functools import partial


convert = NetworkDispatcher('convert')


@convert.register(np.ndarray, pd.DataFrame, cost=0.2)
def dataframe_to_numpy(df, dshape=None, **kwargs):
    dtype = dshape_to_numpy(dshape or discover(df))
    x = df.to_records(index=False)
    if x.dtype != dtype:
        x = x.astype(dtype)
    return x


@convert.register(pd.DataFrame, np.ndarray, cost=1.0)
def numpy_to_dataframe(x, dshape, **kwargs):
    dtype = x.dtype
    names = dtype.names
    if names is None:
        if dtype.kind == 'm':
            # pandas does not do this conversion for us but doesn't work
            # with non 'ns' unit timedeltas
            x = x.astype('m8[ns]')
    else:
        fields = dtype.fields
        new_dtype = []
        should_astype = False
        for name in names:
            original_field_value = fields[name][0]
            if original_field_value.kind == 'm':
                # pandas does not do this conversion for us but doesn't work
                # with non 'ns' unit timedeltas
                new_dtype.append((name, 'm8[ns]'))

                # perform the astype at the end of the loop
                should_astype = True
            else:
                new_dtype.append((name, original_field_value))

        if should_astype:
            x = x.astype(new_dtype)

    df = pd.DataFrame(x, columns=getattr(dshape.measure, 'names', names))
    return df


@convert.register(pd.Series, np.ndarray, cost=1.0)
def numpy_to_series(x, **kwargs):
    names = x.dtype.names
    if names is not None:
        if len(names) > 1:
            raise ValueError('passed in an ndarray with more than 1 column')
        name, = names
        return pd.Series(x[name], name=name)
    return pd.Series(x)


@convert.register(pd.Series, pd.DataFrame, cost=0.1)
def DataFrame_to_Series(x, **kwargs):
    assert len(x.columns) == 1
    return x[x.columns[0]]


@convert.register(pd.DataFrame, pd.Series, cost=0.1)
def series_to_dataframe(x, **kwargs):
    return x.to_frame()


@convert.register(np.recarray, np.ndarray, cost=0.0)
def ndarray_to_recarray(x, **kwargs):
    return x.view(np.recarray)

@convert.register(np.ndarray, np.recarray, cost=0.0)
def recarray_to_ndarray(x, **kwargs):
    return x.view(np.ndarray)


higher_precision_freqs = frozenset(('ns', 'ps', 'fs', 'as'))


@convert.register(np.ndarray, pd.Series, cost=0.1)
def series_to_array(s, dshape=None, **kwargs):
    # if we come from a node that can't be discovered we need to discover
    # on s
    dtype = dshape_to_numpy(datashape.dshape(dshape or discover(s)))
    sdtype = s.dtype
    values = s.values

    # don't lose precision of datetime64 more precise than microseconds
    if ((issubclass(sdtype.type, np.datetime64) and
            np.datetime_data(sdtype)[0] in higher_precision_freqs) or
            s.dtype == dtype):
        return values
    try:
        return values.astype(dtype)
    except ValueError:  # object series and record dshape, e.g., a frame row
        return values


@convert.register(list, np.ndarray, cost=10.0)
def numpy_to_list(x, **kwargs):
    dt = None
    if x.dtype == 'M8[ns]':
        dt = 'M8[us]'  # lose precision when going to Python datetime
    if x.dtype.fields and any(x.dtype[n] == 'M8[ns]' for n in x.dtype.names):
        dt = [(n, 'M8[us]' if x.dtype[n] == 'M8[ns]' else x.dtype[n])
              for n in x.dtype.names]
    if dt:
        return x.astype(dt).tolist()
    else:
        return x.tolist()


@convert.register(np.ndarray, chunks(np.ndarray), cost=1.0)
def numpy_chunks_to_numpy(c, **kwargs):
    return np.concatenate(list(c))


@convert.register(chunks(np.ndarray), np.ndarray, cost=0.5)
def numpy_to_chunks_numpy(x, chunksize=2**20, **kwargs):
    return chunks(np.ndarray)(
            lambda: (x[i:i+chunksize] for i in range(0, x.shape[0], chunksize)))


@convert.register(pd.DataFrame, chunks(pd.DataFrame), cost=1.0)
def chunks_dataframe_to_dataframe(c, **kwargs):
    c = list(c)
    if not c:  # empty case
        return pd.DataFrame(columns=kwargs.get('dshape').measure.names)
    else:
        return pd.concat(c, axis=0, ignore_index=True)


@convert.register(chunks(pd.DataFrame), pd.DataFrame, cost=0.5)
def dataframe_to_chunks_dataframe(x, chunksize=2**20, **kwargs):
    return chunks(pd.DataFrame)(
            lambda: (x.iloc[i:i+chunksize] for i in range(0, x.shape[0], chunksize)))

def ishashable(x):
    try:
        hash(x)
        return True
    except:
        return False


@convert.register(set, (list, tuple), cost=5.0)
def iterable_to_set(x, **kwargs):
    if x and isinstance(x[0], (tuple, list)) and not ishashable(x):
        x = map(tuple, x)
    return set(x)


@convert.register(list, (tuple, set), cost=1.0)
def iterable_to_list(x, **kwargs):
    return list(x)


@convert.register(tuple, (list, set), cost=1.0)
def iterable_to_tuple(x, **kwargs):
    return tuple(x)


def element_of(seq):
    """

    >>> element_of([1, 2, 3])
    1
    >>> element_of([[1, 2], [3, 4]])
    1
    """
    while isinstance(seq, list) and seq:
        seq = seq[0]
    return seq


@convert.register(np.ndarray, list, cost=10.0)
def list_to_numpy(seq, dshape=None, **kwargs):
    if isinstance(element_of(seq), dict):
        seq = list(records_to_tuples(dshape, seq))
    if (seq and isinstance(seq[0], Iterable) and not ishashable(seq[0]) and
            not isscalar(dshape)):
        seq = list(map(tuple, seq))
    return np.array(seq, dtype=dshape_to_numpy(dshape))


@convert.register(Iterator, list, cost=0.001)
def list_to_iterator(L, **kwargs):
    return iter(L)


@convert.register(list, Iterator, cost=1.0)
def iterator_to_list(seq, **kwargs):
    return list(seq)


@convert.register(Iterator, (chunks(pd.DataFrame), chunks(np.ndarray)), cost=10.0)
def numpy_chunks_to_iterator(c, **kwargs):
    return concat(convert(Iterator, chunk, **kwargs) for chunk in c)


@convert.register(chunks(np.ndarray), Iterator, cost=10.0)
def iterator_to_numpy_chunks(seq, chunksize=1024, **kwargs):
    seq2 = partition_all(chunksize, seq)
    try:
        first, rest = next(seq2), seq2
    except StopIteration:  # seq is empty
        def _():
            yield convert(np.ndarray, [], **kwargs)
    else:
        x = convert(np.ndarray, first, **kwargs)

        def _():
            yield x
            for i in rest:
                yield convert(np.ndarray, i, **kwargs)
    return chunks(np.ndarray)(_)


@convert.register(chunks(pd.DataFrame), Iterator, cost=10.0)
def iterator_to_DataFrame_chunks(seq, chunksize=1024, **kwargs):
    seq2 = partition_all(chunksize, seq)

    add_index = kwargs.get('add_index', False)
    if not add_index:
        # Simple, we can dispatch to dask...
        f = lambda d: convert(pd.DataFrame, d, **kwargs)
        data = [partial(f, d) for d in seq2]
        if not data:
            data = [convert(pd.DataFrame, [], **kwargs)]
        return chunks(pd.DataFrame)(data)

    # TODO: Decide whether we should support the `add_index` flag at all.
    # If so, we need to post-process the converted DataFrame objects sequencially,
    # so we can't parallelize the process.
    try:
        first, rest = next(seq2), seq2
    except StopIteration:
        def _():
            yield convert(pd.DataFrame, [], **kwargs)
    else:
        df = convert(pd.DataFrame, first, **kwargs)
        df1, n1 = _add_index(df, 0)

        def _():
            n = n1
            yield df1
            for i in rest:
                df = convert(pd.DataFrame, i, **kwargs)
                df, n = _add_index(df, n)
                yield df
    return chunks(pd.DataFrame)(_)


def _add_index(df, start, _idx_type=getattr(pd, 'RangeIndex',
                                            compose(pd.Index, np.arange))):
    stop = start + len(df)
    idx = _idx_type(start=start, stop=stop)
    df.index = idx
    return df, stop


@convert.register(tuple, np.record)
def numpy_record_to_tuple(rec, **kwargs):
    return rec.tolist()


@convert.register(chunks(np.ndarray), chunks(pd.DataFrame), cost=0.5)
def chunked_pandas_to_chunked_numpy(c, **kwargs):
    return chunks(np.ndarray)(lambda: (convert(np.ndarray, chunk, **kwargs) for chunk in c))

@convert.register(chunks(pd.DataFrame), chunks(np.ndarray), cost=0.5)
def chunked_numpy_to_chunked_pandas(c, **kwargs):
    return chunks(pd.DataFrame)(lambda: (convert(pd.DataFrame, chunk, **kwargs) for chunk in c))


@convert.register(chunks(np.ndarray), chunks(list), cost=10.0)
def chunked_list_to_chunked_numpy(c, **kwargs):
    return chunks(np.ndarray)(lambda: (convert(np.ndarray, chunk, **kwargs) for chunk in c))

@convert.register(chunks(list), chunks(np.ndarray), cost=10.0)
def chunked_numpy_to_chunked_list(c, **kwargs):
    return chunks(list)(lambda: (convert(list, chunk, **kwargs) for chunk in c))

@convert.register(chunks(Iterator), chunks(list), cost=0.1)
def chunked_list_to_chunked_iterator(c, **kwargs):
    return chunks(Iterator)(c.data)

@convert.register(chunks(list), chunks(Iterator), cost=0.1)
def chunked_Iterator_to_chunked_list(c, **kwargs):
    return chunks(Iterator)(lambda: (convert(Iterator, chunk, **kwargs) for chunk in c))

@convert.register(Iterator, chunks(Iterator), cost=0.1)
def chunked_iterator_to_iterator(c, **kwargs):
    return concat(c)


ooc_types |= set([Iterator, Chunks])



from __future__ import absolute_import, division, print_function

import pytest
paramiko = pytest.importorskip('paramiko')

import pandas as pd
import numpy as np
import re
import os
import sys

from odo.utils import tmpfile, filetext
from odo.directory import _Directory, Directory
from odo.backends.ssh import SSH, resource, ssh_pattern, sftp, drop, connect
from odo.backends.csv import CSV
from odo import into, discover, CSV, JSONLines, JSON, convert
from odo.temp import _Temp, Temp
from odo.compatibility import ON_TRAVIS_CI
import socket

skipif = pytest.mark.skipif

# NOTE: this is a workaround for paramiko on Py2; connect() hangs without
# raising an exception.  Shows up on paramiko 1.16.0 and 2.0.2 with Py 2.7.
# KWS: 2016-08-10
# JJ: Still happens as of 2016-10-20
try_to_connect = sys.version_info[0] >= 3
pytestmark = skipif(not try_to_connect, reason='could not connect')

if try_to_connect:
    try:
        ssh = connect(hostname='localhost')
        ssh.close()
    except socket.error:
        pytestmark = pytest.mark.skip('Could not connect')
    except paramiko.PasswordRequiredException as e:
        pytestmark = pytest.mark.skip(str(e))
    except paramiko.SSHException as e:
        pytestmark = pytest.mark.skip(str(e))
    except TypeError:
        # NOTE: This is a workaround for paramiko version 1.16.0 on Python 3.4,
        # that raises a TypeError due to improper indexing internally into
        # dict_keys when a ConnectionRefused error is raised.
        # KWS 2016-04-21.
        pytestmark = pytest.mark.skip('Could not connect')


def test_resource():
    r = resource('ssh://joe@localhost:/path/to/myfile.csv')
    assert isinstance(r, SSH(CSV))
    assert r.path == '/path/to/myfile.csv'
    assert r.auth['hostname'] == 'localhost'
    assert r.auth['username'] == 'joe'


def test_connect():
    a = connect(hostname='localhost')
    b = connect(hostname='localhost')
    assert a is b

    a.close()

    c = connect(hostname='localhost')
    assert a is c
    assert c.get_transport() and c.get_transport().is_active()


def test_resource_directory():
    r = resource('ssh://joe@localhost:/path/to/')
    assert issubclass(r.subtype, _Directory)

    r = resource('ssh://joe@localhost:/path/to/*.csv')
    assert r.subtype == Directory(CSV)
    assert r.path == '/path/to/'


def test_discover():
    with filetext('name,balance\nAlice,100\nBob,200') as fn:
        local = CSV(fn)
        remote = SSH(CSV)(fn, hostname='localhost')

        assert discover(local) == discover(remote)


def test_discover_from_resource():
    with filetext('name,balance\nAlice,100\nBob,200', extension='csv') as fn:
        local = CSV(fn)
        remote = resource('ssh://localhost:' + fn)

        assert discover(local) == discover(remote)


def test_ssh_pattern():
    uris = ['localhost:myfile.csv',
            '127.0.0.1:/myfile.csv',
            'user@127.0.0.1:/myfile.csv',
            'user@127.0.0.1:/*.csv',
            'user@127.0.0.1:/my-dir/my-file3.csv']
    for uri in uris:
        assert re.match(ssh_pattern, uri)


def test_copy_remote_csv():
    with tmpfile('csv') as target:
        with filetext('name,balance\nAlice,100\nBob,200',
                      extension='csv') as fn:
            csv = resource(fn)

            uri = 'ssh://localhost:%s.csv' % target
            scsv = into(uri, csv)

            assert isinstance(scsv, SSH(CSV))
            assert discover(scsv) == discover(csv)

            # Round trip
            csv2 = into(target, scsv)
            assert into(list, csv) == into(list, csv2)


def test_drop():
    with filetext('name,balance\nAlice,100\nBob,200', extension='csv') as fn:
        with tmpfile('csv') as target:
            scsv = SSH(CSV)(target, hostname='localhost')

            assert not os.path.exists(target)

            conn = sftp(**scsv.auth)
            conn.put(fn, target)

            assert os.path.exists(target)

            drop(scsv)
            drop(scsv)

            assert not os.path.exists(target)


def test_drop_of_csv_json_lines_use_ssh_version():
    from odo.backends.ssh import drop_ssh
    for typ in [CSV, JSON, JSONLines]:
        assert drop.dispatch(SSH(typ)) == drop_ssh


def test_convert_local_file_to_temp_ssh_file():
    with filetext('name,balance\nAlice,100\nBob,200', extension='csv') as fn:
        csv = CSV(fn)
        scsv = convert(Temp(SSH(CSV)), csv, hostname='localhost')

        assert into(list, csv) == into(list, scsv)


@skipif(ON_TRAVIS_CI, reason="Don't know")
def test_temp_ssh_files():
    with filetext('name,balance\nAlice,100\nBob,200', extension='csv') as fn:
        csv = CSV(fn)
        scsv = into(Temp(SSH(CSV)), csv, hostname='localhost')
        assert discover(csv) == discover(scsv)

        assert isinstance(scsv, _Temp)


@skipif(ON_TRAVIS_CI, reason="Don't know")
def test_convert_through_temporary_local_storage():
    with filetext('name,quantity\nAlice,100\nBob,200', extension='csv') as fn:
        csv = CSV(fn)
        df = into(pd.DataFrame, csv)
        scsv = into(Temp(SSH(CSV)), csv, hostname='localhost')

        assert into(list, csv) == into(list, scsv)

        scsv2 = into(Temp(SSH(CSV)), df, hostname='localhost')
        assert into(list, scsv2) == into(list, df)

        sjson = into(Temp(SSH(JSONLines)), df, hostname='localhost')
        assert (into(np.ndarray, sjson) == into(np.ndarray, df)).all()


@skipif(ON_TRAVIS_CI and sys.version_info[0] == 3,
        reason='Strange hanging on travis for python33 and python34')
def test_ssh_csv_to_s3_csv():
    # for some reason this can only be run in the same file as other ssh tests
    # and must be a Temp(SSH(CSV)) otherwise tests above this one fail
    s3_bucket = pytest.importorskip('odo.backends.tests.test_aws').s3_bucket

    with filetext('name,balance\nAlice,100\nBob,200', extension='csv') as fn:
        remote = into(Temp(SSH(CSV)), CSV(fn), hostname='localhost')
        with s3_bucket('.csv') as b:
            result = into(b, remote)
            assert discover(result) == discover(resource(b))


@skipif(ON_TRAVIS_CI and sys.version_info[0] == 3,
        reason='Strange hanging on travis for python33 and python34')
def test_s3_to_ssh():
    pytest.importorskip('boto')

    tips_uri = 's3://nyqpug/tips.csv'
    with tmpfile('.csv') as fn:
        result = into(Temp(SSH(CSV))(fn, hostname='localhost'), tips_uri)
        assert into(list, result) == into(list, tips_uri)
        assert discover(result) == discover(resource(tips_uri))



from __future__ import absolute_import, division, print_function

import os
from itertools import product
import pandas as pd
import pytest
sa = pytest.importorskip('sqlalchemy')

from datashape import dshape, discover
from odo import resource, odo
from odo.utils import tmpfile, filetext


ds = dshape('var *  {a: int32, b: int32}')
data = [(1, 2), (10, 20), (100, 200)]


@pytest.yield_fixture
def csv():
    with tmpfile('csv') as filename:
        yield odo(data, filename, dshape=ds, has_header=False)


def test_simple_into(csv):
    tbl = 'testtable'
    with tmpfile('db') as filename:
        engine = resource('sqlite:///' + filename)
        t = resource('sqlite:///' + filename + '::' + tbl,
                     dshape=ds)

        odo(csv, t, dshape=ds)
        conn = engine.raw_connection()
        cursor = conn.cursor()
        cursor.execute("""SELECT
                        name
                       FROM
                        sqlite_master
                       WHERE type='table' and name='{0}';""".format(tbl))

        sqlite_tbl_names = cursor.fetchall()
        conn.close()

        assert sqlite_tbl_names[0][0] == tbl
        assert odo(t, list) == data


def test_csv_with_header():
    with tmpfile('db') as dbfilename:
        with filetext('a,b\n1,2\n3,4', extension='csv') as csvfilename:
            t = odo(csvfilename, 'sqlite:///%s::mytable' % dbfilename)
            assert discover(t) == dshape('var * {a: int64, b: int64}')
            assert odo(t, set) == set([(1, 2), (3, 4)])


def test_csv_infer_header():
    with tmpfile('db') as dbfilename:
        with filetext('a,b\n1,2\n3,4', extension='csv') as csvfilename:
            t = odo(csvfilename, 'sqlite:///%s::mytable' % dbfilename)
            assert discover(t) == dshape('var * {a: int64, b: int64}')
            assert odo(t, set) == set([(1, 2), (3, 4)])


@pytest.mark.parametrize(['sep', 'header'],
                         product([',', '|', '\t'], [True, False]))
def test_sqlite_to_csv(sep, header):
    with tmpfile('db') as dbfilename:
        with filetext('a,b\n1,2\n3,4', extension='csv') as csvfilename:
            t = odo(csvfilename, 'sqlite:///%s::mytable' % dbfilename)

        with tmpfile('.csv') as fn:
            odo(t, fn, header=header, delimiter=sep)
            with open(fn, 'rt') as f:
                lines = f.readlines()
            expected = [tuple(map(int, row))
                        for row in map(lambda x: x.split(sep), lines[header:])]
            assert odo(fn, list, delimiter=sep, has_header=header,
                       dshape=discover(t)) == expected


def test_different_encoding():
    encoding = 'latin1'
    with tmpfile('db') as db:
        sql = odo(os.path.join(os.path.dirname(__file__), 'encoding.csv'),
                  'sqlite:///%s::t' % db, encoding=encoding)
        result = odo(sql, list)

    NULL = u''

    expected = [(u'1958.001.500131-1A', 1, NULL, NULL, 899),
                (u'1958.001.500156-6', 1, NULL, NULL, 899),
                (u'1958.001.500162-1', 1, NULL, NULL, 899),
                (u'1958.001.500204-2', 1, NULL, NULL, 899),
                (u'1958.001.500204-2A', 1, NULL, NULL, 899),
                (u'1958.001.500204-2B', 1, NULL, NULL, 899),
                (u'1958.001.500223-6', 1, NULL, NULL, 9610),
                (u'1958.001.500233-9', 1, NULL, NULL, 4703),
                (u'1909.017.000018-3', 1, 30.0, u'sumaria', 899)]
    assert result == expected


@pytest.yield_fixture
def quoted_sql():
    with tmpfile('.db') as db:
        try:
            yield resource('sqlite:///%s::foo bar' % db, dshape=ds)
        except sa.exc.OperationalError as e:
            pytest.skip(str(e))


@pytest.mark.xfail(
    raises=sa.exc.DatabaseError,
    reason='How do you use a quoted table name with the SQLite .import command?'
)
def test_quoted_name(csv, quoted_sql):
    with tmpfile('csv') as filename:
        csv = odo(data, filename, dshape=ds, has_header=True)
        s = odo(csv, quoted_sql)
        t = odo(csv, list)
        assert sorted(odo(s, list)) == sorted(t)


def test_different_encoding_to_csv():
    with tmpfile('db') as dbfilename:
        with filetext('a,b\n1,2\n3,4', extension='csv') as csvfilename:
            t = odo(
                csvfilename,
                'sqlite:///%s::mytable' % dbfilename,
                encoding='latin1'
            )
            with tmpfile('.csv') as fn:
                with pytest.raises(ValueError):
                    odo(t, fn, encoding='latin1')


def test_send_parameterized_query_to_csv():
    with tmpfile('db') as dbfilename:
        with filetext('a,b\n1,2\n3,4', extension='csv') as csvfilename:
            t = odo(
                csvfilename,
                'sqlite:///%s::mytable' % dbfilename,
            )
        with tmpfile('.csv') as fn:
            q = t.select(t.c.a == 1)
            r = odo(q, fn)
            assert sorted(odo(q, list)) == sorted(odo(r, list))


def test_df_to_in_memory_db():
    df = pd.DataFrame([[1, 2], [3, 4]], columns=list('ab'))
    tbl = odo(df, 'sqlite:///:memory:::tbl')
    pd.util.testing.assert_frame_equal(
        odo(tbl, pd.DataFrame),
        df,
    )



from __future__ import absolute_import, division, print_function

import paramiko
from contextlib import contextmanager
from toolz import keyfilter, memoize, take, curry
from datashape import discover
import re
import uuid

from ..directory import Directory
from ..utils import keywords, tmpfile, sample, ignoring, copydoc
from ..resource import resource
from ..append import append
from ..convert import convert
from ..temp import Temp, _Temp
from ..drop import drop
from .csv import CSV
from .json import JSON, JSONLines
from .text import TextFile

connection_pool = dict()

def connect(**auth):
    key = tuple(sorted(auth.items()))
    if key in connection_pool:
        ssh = connection_pool[key]
        if not ssh.get_transport() or not ssh.get_transport().is_active():
            ssh.connect(**auth)
    else:
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(**auth)
        connection_pool[key] = ssh
    return ssh

sftp_pool = dict()

def sftp(**auth):
    ssh = connect(**auth)  # Need to call this explicitly (can't memoize)
    key = tuple(sorted(auth.items()))
    if key in sftp_pool:
        conn = sftp_pool[key]
    else:
        conn = ssh.open_sftp()
        sftp_pool[key] = conn

    conn.sock.setblocking(True)

    return conn


class _SSH(object):
    """ Parent class for data accessed through ``ssh``

    See ``paramiko.SSHClient.connect`` for authentication keyword arguments

    Examples
    --------

    >>> from odo import SSH, CSV
    >>> s = SSH(CSV)('/path/to/file.csv', hostname='hostname', username='alice')

    Normally create through resource uris

    >>> data = resource('ssh://alice@host:/path/to/file.csv', password='pass')
    >>> data.path
    '/path/to/file.csv'
    >>> data.auth['hostname']
    'host'
    """
    def __init__(self, *args, **kwargs):
        self.auth = keyfilter(keywords(paramiko.SSHClient.connect).__contains__,
                              kwargs)
        self.subtype.__init__(self, *args, **kwargs)

    def lines(self):
        conn = sftp(**self.auth)
        return conn.file(self.path, 'r')


@memoize
@copydoc(_SSH)
def SSH(cls):
    return type('SSH(%s)' % cls.__name__, (_SSH, cls), {'subtype':  cls})


types_by_extension = {'csv': CSV, 'json': JSONLines}

ssh_pattern = '((?P<username>[a-zA-Z]\w*)@)?(?P<hostname>[\w.-]*)(:(?P<port>\d+))?:(?P<path>[/\w.*-]+)'

@resource.register('ssh://.+', priority=16)
def resource_ssh(uri, **kwargs):
    if 'ssh://' in uri:
        uri = uri[len('ssh://'):]

    d = re.match(ssh_pattern, uri).groupdict()
    d = dict((k, v) for k, v in d.items() if v is not None)
    path = d.pop('path')

    kwargs.update(d)

    try:
        subtype = types_by_extension[path.split('.')[-1]]
        if '*' in path:
            subtype = Directory(subtype)
            path = path.rsplit('/', 1)[0] + '/'
    except KeyError:
        subtype = type(resource(path))

    return SSH(subtype)(path, **kwargs)


@sample.register((SSH(CSV),
                  SSH(JSON),
                  SSH(JSONLines)))
@contextmanager
def sample_ssh(data, lines=500):
    """ Grab a few lines from the remote file """
    with tmpfile() as fn:
        with open(fn, 'w') as f:
            for line in take(lines, data.lines()):
                f.write(line)
        yield fn


@sample.register((SSH(Directory(CSV)),
                  SSH(Directory(JSON)),
                  SSH(Directory(JSONLines))))
@contextmanager
def sample_ssh(data, **kwargs):
    """ Grab a few lines from a file in a remote directory """
    conn = sftp(**data.auth)
    fn = data.path + '/' + conn.listdir(data.path)[0]
    one_file = SSH(data.container)(fn, **data.auth)
    with sample(one_file, **kwargs) as result:
        yield result


@discover.register(_SSH)
def discover_ssh(data, **kwargs):
    with sample(data) as fn:
        o = data.subtype(fn)
        result = discover(o)
    return result


@discover.register(SSH(CSV))
def discover_ssh_csv(data, **kwargs):
    with sample(data) as fn:
        o = CSV(fn, encoding=data.encoding, has_header=data.has_header, **data.dialect)
        result = discover(o)
    return result


@discover.register((SSH(JSON), SSH(JSONLines)))
def discover_ssh_json(data, **kwargs):
    with sample(data) as fn:
        result = discover(data.subtype(fn))
    return result


@discover.register((SSH(Directory(CSV)),
                    SSH(Directory(JSON)),
                    SSH(Directory(JSONLines))))
def discover_ssh_directory(data, **kwargs):
    conn = sftp(**data.auth)
    fn = data.path + '/' + conn.listdir(data.path)[0]
    one_file = SSH(data.container)(fn, **data.auth)
    result = discover(one_file)
    return result


@drop.register((_SSH, SSH(CSV), SSH(JSON), SSH(JSONLines)))
def drop_ssh(data, **kwargs):
    conn = sftp(**data.auth)
    with ignoring(IOError):
        conn.remove(data.path)


@append.register(_SSH, object)
def append_anything_to_ssh(target, source, **kwargs):
    if not isinstance(source, target.subtype):
        source = convert(Temp(target.subtype), source, **kwargs)
    # TODO: handle overwrite case
    conn = sftp(**target.auth)
    conn.put(source.path, target.path)
    return target


@append.register(TextFile, SSH(TextFile))
@append.register(JSONLines, SSH(JSONLines))
@append.register(JSON, SSH(JSON))
@append.register(CSV, SSH(CSV))
def append_sshX_to_X(target, source, **kwargs):
    # TODO: handle overwrite case
    conn = sftp(**source.auth)
    conn.get(source.path, target.path)
    return target


@curry
def file_to_temp_ssh_file(typ, data, **kwargs):
    """ Generic convert function sending data to ssh(data)

    Needs to be partially evaluated with a type"""
    # don't use . prefix to hide because Hive doesn't like it
    fn = '%s.%s' % (uuid.uuid1(), typ.canonical_extension)
    target = Temp(SSH(typ))(fn, **kwargs)
    return append(target, data, **kwargs)

for typ in [CSV, JSON, JSONLines, TextFile]:
    convert.register(Temp(SSH(typ)), (Temp(typ), typ))(
            file_to_temp_ssh_file(typ))


@convert.register(Temp(TextFile), (Temp(SSH(TextFile)), SSH(TextFile)))
@convert.register(Temp(JSONLines), (Temp(SSH(JSONLines)), SSH(JSONLines)))
@convert.register(Temp(JSON), (Temp(SSH(JSON)), SSH(JSON)))
@convert.register(Temp(CSV), (Temp(SSH(CSV)), SSH(CSV)))
def ssh_file_to_temp_file(data, **kwargs):
    fn = '.%s' % uuid.uuid1()
    target = Temp(data.subtype)(fn, **kwargs)
    return append(target, data, **kwargs)


@convert.register(Temp(SSH(TextFile)), (TextFile, Temp(TextFile)))
@convert.register(Temp(SSH(JSONLines)), (JSONLines, Temp(JSONLines)))
@convert.register(Temp(SSH(JSON)), (JSON, Temp(JSON)))
@convert.register(Temp(SSH(CSV)), (CSV, Temp(CSV)))
def file_to_temp_ssh_file(data, **kwargs):
    fn = '%s' % uuid.uuid1()
    if isinstance(data, _Temp):
        target = Temp(SSH(data.persistent_type))(fn, **kwargs)
    else:
        target = Temp(SSH(type(data)))(fn, **kwargs)
    return append(target, data, **kwargs)



from __future__ import absolute_import, division, print_function

from odo.backends.text import (TextFile, resource, convert, discover, append,
        drop, chunks, Temp)
from odo.utils import tmpfile, filetexts, filetext
from datashape import dshape
import gzip
import os


def test_resource():
    assert isinstance(resource('foo.txt'), TextFile)
    assert isinstance(resource('foo.log.gz'), TextFile)
    assert isinstance(resource('/path/to/foo.log'), TextFile)


def test_open():
    r = resource('foo.log.gz')
    assert r.open == gzip.open


def test_convert():
    with filetext('Hello\nWorld') as fn:
        assert convert(list, TextFile(fn)) == ['Hello\n', 'World']


def test_append():
    with tmpfile('log') as fn:
        t = TextFile(fn)
        append(t, ['Hello', 'World'])

        assert os.path.exists(fn)
        with open(fn) as f:
            assert list(map(str.strip, f.readlines())) == ['Hello', 'World']


def test_discover():
    assert discover(TextFile('')) == dshape('var * string')


def test_drop():
    with filetext('hello\nworld') as fn:
        t = TextFile(fn)
        assert os.path.exists(fn)
        drop(t)
        assert not os.path.exists(fn)


def test_chunks_textfile():
    with filetexts({'a1.log': 'Hello\nWorld', 'a2.log': 'Hola\nMundo'}) as fns:
        logs = chunks(TextFile)(list(map(TextFile, fns)))
        assert set(map(str.strip, convert(list, logs))) == \
                set(['Hello', 'World', 'Hola', 'Mundo'])

def test_temp():
    t = convert(Temp(TextFile), [1, 2, 3])
    assert [int(line.strip()) for line in convert(list, t)] == [1, 2, 3]



from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd

import datashape
from datashape import discover
from ..append import append
from ..convert import convert, ooc_types
from ..chunks import chunks
from ..resource import resource
from ..utils import filter_kwargs


@discover.register(pd.HDFStore)
def discover_hdfstore(f):
    d = dict()
    for key in f.keys():
        d2 = d
        key2 = key.lstrip('/')
        while '/' in key2:
            group, key2 = key2.split('/', 1)
            if group not in d2:
                d2[group] = dict()
            d2 = d2[group]
        d2[key2] = f.get_storer(key)
    return discover(d)


@discover.register(pd.io.pytables.Fixed)
def discover_hdfstore_storer(storer):
    f = storer.parent
    n = storer.shape
    if isinstance(n, list):
        n = n[0]
    measure = discover(f.select(storer.pathname, start=0, stop=10)).measure
    return n * measure


@convert.register(chunks(pd.DataFrame), pd.io.pytables.AppendableFrameTable)
def hdfstore_to_chunks_dataframes(data, chunksize=100000, **kwargs):
    if (isinstance(chunksize, (float, np.floating)) and
            not chunksize.is_integer()):
        raise TypeError('chunksize argument must be an integer, got %s' %
                        chunksize)

    chunksize = int(chunksize)

    def f():
        k = min(chunksize, 100)
        yield data.parent.select(data.pathname, start=0, stop=k)
        for chunk in data.parent.select(data.pathname, chunksize=chunksize,
                                        start=k):
            yield chunk
    return chunks(pd.DataFrame)(f)


@convert.register(pd.DataFrame, (pd.io.pytables.AppendableFrameTable,
                                 pd.io.pytables.FrameFixed))
def hdfstore_to_chunks_dataframes(data, **kwargs):
    return data.read()


pytables_h5py_explanation = """
You've run in to a conflict between the two HDF5 libraries in Python,
H5Py and PyTables.  You're trying to do something that requires PyTables but
H5Py was loaded first and the two libraries don't share well.

To resolve this you'll have to restart your Python process and ensure that you

    import tables

before you import projects like odo or into or blaze."""

from collections import namedtuple

EmptyHDFStoreDataset = namedtuple('EmptyHDFStoreDataset', 'parent,pathname,dshape')

@resource.register('hdfstore://.+', priority=11)
def resource_hdfstore(uri, datapath=None, dshape=None, **kwargs):
    # TODO:
    # 1. Support nested datashapes (e.g. groups)
    # 2. Try translating unicode to ascii?  (PyTables fails here)
    fn = uri.split('://')[1]
    try:
        f = pd.HDFStore(fn, **filter_kwargs(pd.HDFStore, kwargs))
    except RuntimeError as e:
        raise type(e)(pytables_h5py_explanation)

    if dshape is None:
        return f.get_storer(datapath) if datapath else f
    dshape = datashape.dshape(dshape)

    # Already exists, return it
    if datapath in f:
        return f.get_storer(datapath)

    # Need to create new dataset.
    # HDFStore doesn't support empty datasets, so we use a proxy object.
    return EmptyHDFStoreDataset(f, datapath, dshape)


@append.register((pd.io.pytables.Fixed, EmptyHDFStoreDataset), pd.DataFrame)
def append_dataframe_to_hdfstore(store, df, **kwargs):
    store.parent.append(store.pathname, df, append=True)
    return store.parent.get_storer(store.pathname)


@append.register((pd.io.pytables.Fixed, EmptyHDFStoreDataset),
                 chunks(pd.DataFrame))
def append_chunks_dataframe_to_hdfstore(store, c, **kwargs):
    parent = store.parent
    for chunk in c:
        parent.append(store.pathname, chunk)
    return parent.get_storer(store.pathname)


@append.register((pd.io.pytables.Fixed, EmptyHDFStoreDataset), object)
def append_object_to_hdfstore(store, o, **kwargs):
    return append(store, convert(chunks(pd.DataFrame), o, **kwargs), **kwargs)


ooc_types.add(pd.io.pytables.AppendableFrameTable)



import pytest

pytest.importorskip('dask.bag')

from operator import methodcaller

from odo import chunks, TextFile, odo
from dask.bag import Bag
from odo.utils import filetexts


def inc(x):
    return x + 1


dsk = {('x', 0): (range, 5),
       ('x', 1): (range, 5),
       ('x', 2): (range, 5)}

L = list(range(5)) * 3

b = Bag(dsk, 'x', 3)


def test_convert_bag_to_list():
    assert odo(b, list) == L


def test_convert_logfiles_to_bag():
    with filetexts({'a1.log': 'Hello\nWorld', 'a2.log': 'Hola\nMundo'}) as fns:
        logs = chunks(TextFile)(list(map(TextFile, fns)))
        b = odo(logs, Bag)
        assert isinstance(b, Bag)
        assert 'a1.log' in str(b.dask.values())
        assert (list(map(methodcaller('strip'), odo(b, list))) ==
                list(map(methodcaller('strip'), odo(logs, list))))


def test_sequence():
    b = odo([1, 2, 3], Bag)
    assert set(b.map(inc)) == set([2, 3, 4])



from __future__ import absolute_import, division, print_function

from datashape import dshape
from odo.chunks import *
from toolz import first


CL = chunks(list)


def test_chunks_basics():
    assert isinstance(CL, type)
    assert issubclass(CL, Chunks)


def test_chunks_isnt_consumable():
    cl = CL([[1, 2, 3], [4, 5, 6]])

    assert next(iter(cl)) == [1, 2, 3]
    assert next(iter(cl)) == [1, 2, 3]


def test_chunks_is_memoized():
    assert chunks(list) is chunks(list)


def test_callables():
    cl = CL(lambda: (list(range(3)) for i in range(3)))

    assert first(cl) == [0, 1, 2]
    assert first(cl) == [0, 1, 2]


def test_discover():
    cl = CL([[1, 2, 3], [4, 5, 6]])
    assert discover(cl).measure == discover(1).measure


def test_discover_no_consume():
    cl = CL(iter(([0, 1, 2], [3, 4, 5])))
    assert discover(cl) == discover(cl) == dshape('var * int64')
    assert tuple(cl) == ([0, 1, 2], [3, 4, 5])
    assert tuple(cl) == ()



from __future__ import absolute_import, division, print_function

import pymongo
from pymongo.collection import Collection
from collections import Iterator
from datashape import discover, DataShape, Record, var
from datashape.predicates import isdimension
from datashape.dispatch import dispatch
from toolz import take, partition_all, concat, pluck
import copy
from bson.objectid import ObjectId
import re
from ..convert import convert, ooc_types
from ..append import append
from ..resource import resource


@discover.register(Collection)
def discover_pymongo_collection(coll, n=50):
    items = list(take(n, coll.find()))
    if not items:
        return var * Record([])
    oid_cols = [k for k, v in items[0].items() if isinstance(v, ObjectId)]
    for item in items:
        for col in oid_cols:
            del item[col]

    ds = discover(items)

    if isdimension(ds[0]):
        return coll.count() * ds.subshape[0]
    else:
        raise ValueError("Consistent datashape not found")


def _into_iter_mongodb(coll, columns=None, dshape=None):
    """ Into helper function

    Return both a lazy sequence of tuples and a list of column names
    """
    seq = coll.find()
    if not columns and dshape:
        columns = dshape.measure.names
    elif not columns:
        item = next(seq)
        seq = concat([[item], seq])
        columns = sorted(item.keys())
        columns.remove('_id')
    return columns, pluck(columns, seq)


@convert.register(Iterator, Collection, cost=500.0)
def collection_to_iterator(coll, columns=None, dshape=None, **kwargs):
    columns, seq = _into_iter_mongodb(coll, columns=columns, dshape=dshape)
    return seq


@append.register(Collection, Iterator)
def append_iterator_to_pymongo(coll, seq, columns=None, dshape=None, chunksize=1024, **kwargs):
    seq = iter(seq)
    item = next(seq)
    seq = concat([[item], seq])

    if isinstance(item, (tuple, list)):
        if not columns and dshape:
            columns = dshape.measure.names
        if not columns:
            raise ValueError("Inputs must be dictionaries. "
                "Or provide columns=[...] or dshape=DataShape(...) keyword")
        seq = (dict(zip(columns, item)) for item in seq)

    for block in partition_all(1024, seq):
        coll.insert(copy.deepcopy(block))

    return coll


@append.register(Collection, object)
def append_anything_to_collection(coll, o, **kwargs):
    return append(coll, convert(Iterator, o, **kwargs), **kwargs)


@resource.register(r'mongodb://\w*:\w*@\w*.*', priority=11)
def resource_mongo_with_authentication(uri, collection_name=None, **kwargs):
    pattern = r'mongodb://(?P<user>\w*):(?P<pass>\w*)@(?P<hostport>.*:?\d*)/(?P<database>\w+)'
    d = re.search(pattern, uri).groupdict()
    return _resource_mongo(d, collection_name)


@resource.register(r'mongodb://.+')
def resource_mongo(uri, collection_name=None, **kwargs):
    pattern = r'mongodb://(?P<hostport>.*:?\d*)/(?P<database>\w+)'
    d = re.search(pattern, uri).groupdict()
    return _resource_mongo(d, collection_name)


def _resource_mongo(d, collection_name=None):
    client = pymongo.MongoClient(d['hostport'])
    db = getattr(client, d['database'])
    if d.get('user'):
        db.authenticate(d['user'], d['pass'])
    if collection_name is None:
        return db
    return getattr(db, collection_name)


@discover.register(pymongo.database.Database)
def discover_mongo_database(db):
    names = db.collection_names()
    return DataShape(Record(zip(names, (discover(getattr(db, name))
                                        for name in names))))


ooc_types.add(Collection)


@dispatch(Collection)
def drop(m):
    m.drop()



from __future__ import absolute_import, division, print_function

from collections import OrderedDict
from datetime import datetime, timedelta
from distutils.version import LooseVersion

from datashape import (
    Categorical,
    DateTime,
    Option,
    Record,
    discover,
    dshape,
    float64,
    int64,
)
from datashape.util.testing import assert_dshape_equal
from networkx import NetworkXNoPath
import numpy as np
import pandas as pd
import pytest

from odo import odo


data = [('Alice', 100), ('Bob', 200)]


def test_discover_dataframe():
    df = pd.DataFrame([('Alice', 100), ('Bob', 200)],
                      columns=['name', 'amount'])

    assert discover(df) == dshape('2 * {name: ?string, amount: int64}')


def test_discover_series():
    s = pd.Series([1, 2, 3])

    assert discover(s) == 3 * discover(s[0])


def test_floats_are_not_optional():
    df = pd.DataFrame([('Alice', 100), ('Bob', None)],
                      columns=['name', 'amount'])
    ds = discover(df)
    assert_dshape_equal(ds[1].types[1], float64)


def test_datetime_to_timestamp():
    dt = datetime(2014, 1, 1)
    ts = odo(dt, pd.Timestamp)
    assert isinstance(ts, pd.Timestamp)
    assert ts == pd.Timestamp('2014-01-01')


def test_nan_to_nat():
    assert odo(float('nan'), pd.Timestamp) is pd.NaT
    assert odo(float('nan'), pd.Timedelta) is pd.NaT
    assert odo(np.nan, pd.Timestamp) is pd.NaT
    assert odo(np.nan, pd.Timedelta) is pd.NaT

    with pytest.raises(NetworkXNoPath):
        # Check that only nan can be converted.
        odo(0.5, pd.Timestamp)

    with pytest.raises(NetworkXNoPath):
        # Check that only nan can be converted.
        odo(0.5, pd.Timedelta)


def test_none_to_nat():
    assert odo(None, pd.Timestamp) is pd.NaT
    assert odo(None, pd.Timedelta) is pd.NaT


def test_nat_to_nat():
    assert odo(pd.NaT, pd.Timestamp) is pd.NaT
    assert odo(pd.NaT, pd.Timedelta) is pd.NaT


def test_timedelta_to_pandas():
    assert odo(timedelta(days=1), pd.Timedelta) == pd.Timedelta(days=1)
    assert odo(timedelta(hours=1), pd.Timedelta) == pd.Timedelta(hours=1)
    assert odo(timedelta(seconds=1), pd.Timedelta) == pd.Timedelta(seconds=1)


def test_categorical_pandas():
    df = pd.DataFrame({'x': list('a'*5 + 'b'*5 + 'c'*5),
                       'y': np.arange(15, dtype=np.int64)},
		      columns=['x', 'y'])
    df.x = df.x.astype('category')
    assert_dshape_equal(discover(df), 15 * Record([('x',
                        Categorical(['a', 'b', 'c'])), ('y', int64)]))
    assert_dshape_equal(discover(df.x), 15 * Categorical(['a', 'b', 'c']))


@pytest.mark.skipif(
    LooseVersion(pd.__version__) < LooseVersion('0.17'),
    reason="Pandas before DatetimeTZ",
)
def test_datetimetz_pandas():
    df = pd.DataFrame(
        OrderedDict([
            ('naive', pd.date_range('2014', periods=5)),
            ('Europe/Moscow', pd.date_range('2014', periods=5, tz='Europe/Moscow')),
            ('UTC', pd.date_range('2014', periods=5, tz='UTC')),
            ('US/Eastern', pd.date_range('2014', periods=5, tz='US/Eastern')),
        ])
    )

    assert_dshape_equal(
        discover(df),
        5 * Record[
            'naive': Option(DateTime(tz=None)),
            'Europe/Moscow': Option(DateTime(tz='Europe/Moscow')),
            'UTC': Option(DateTime(tz='UTC')),
            'US/Eastern': Option(DateTime(tz='US/Eastern')),
        ]
    )

    assert_dshape_equal(discover(df.naive), 5 * Option(DateTime(tz=None)))
    for tz in ('Europe/Moscow', 'UTC', 'US/Eastern'):
        assert_dshape_equal(
            discover(df[tz]),
            5 * Option(DateTime(tz=tz))
        )



from __future__ import absolute_import, division, print_function

import gzip
from datashape import discover, dshape
from collections import Iterator
from toolz import partial, concat
import uuid
import os

from ..compatibility import unicode
from ..chunks import chunks
from ..drop import drop
from ..temp import Temp
from ..append import append
from ..convert import convert
from ..resource import resource


class TextFile(object):
    canonical_extension = 'txt'

    def __init__(self, path, **kwargs):
        self.path = path

    @property
    def open(self):
        if self.path.split(os.path.extsep)[-1] == 'gz':
            return gzip.open
        else:
            return open


@convert.register(Iterator, (TextFile, Temp(TextFile)), cost=0.1)
def textfile_to_iterator(data, **kwargs):
    with data.open(data.path) as f:
        for line in f:
            yield line


@convert.register(Iterator, chunks(TextFile), cost=0.1)
def chunks_textfile_to_iterator(data, **kwargs):
    return concat(map(partial(convert, Iterator), data))


@discover.register((TextFile, Temp(TextFile)))
def discover_textfile(data, **kwargs):
    return dshape('var * string')


@append.register((Temp(TextFile), TextFile), Iterator)
def append_iterator_to_textfile(target, source, **kwargs):
    with target.open(target.path, 'a') as f:
        for item in source:
            f.write(unicode(item))
            f.write('\n')  # TODO: detect OS-level newline character
    return target


@append.register(TextFile, object)
def append_anything_to_textfile(target, source, **kwargs):
    return append(target, convert(Iterator, source, **kwargs), **kwargs)


@convert.register(Temp(TextFile), Iterator)
def iterator_to_temp_textfile(seq, **kwargs):
    fn = str(uuid.uuid1())
    txt = Temp(TextFile)(fn)
    return append(txt, seq, **kwargs)


@resource.register('.+\.(txt|log)(.gz)?')
def resource_sas(uri, **kwargs):
    return TextFile(uri)


@drop.register(TextFile)
def drop_textfile(data, **kwargs):
    os.remove(data.path)



from __future__ import absolute_import, division, print_function
import warnings

from odo.core import NetworkDispatcher, path, FailedConversionWarning
from datashape import discover

d = NetworkDispatcher('foo')

@d.register(float, int, cost=1.0)
def f(x, **kwargs):
    return float(x)

@d.register(str, float, cost=1.0)
def g(x, **kwargs):
    return str(x)


def test_basic():
    assert [func for a, b, func in d.path(int, str)] == [f, g]

    assert d.path(int, str) == d.path(1, '')


def test_convert_is_robust_to_failures():
    foo = NetworkDispatcher('foo')

    def badfunc(*args, **kwargs):
        raise NotImplementedError()

    class A(object): pass
    class B(object): pass
    class C(object): pass
    discover.register((A, B, C))(lambda x: 'int')
    foo.register(B, A, cost=1.0)(lambda x, **kwargs: 1)
    foo.register(C, B, cost=1.0)(badfunc)
    foo.register(C, A, cost=10.0)(lambda x, **kwargs: 2)

    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter('always')
        assert foo(C, A()) == 2

    assert len(ws) == 1
    w = ws[0].message
    assert isinstance(w, FailedConversionWarning)
    assert 'B -> C' in str(w)


def test_ooc_behavior():
    foo = NetworkDispatcher('foo')
    class A(object): pass
    class B(object): pass
    class C(object): pass

    discover.register((A, B, C))(lambda x: 'int')
    foo.register(B, A, cost=1.0)(lambda x, **kwargs: 1)
    foo.register(C, B, cost=1.0)(lambda x, **kwargs: x / 0) # note that this errs
    foo.register(C, A, cost=10.0)(lambda x, **kwargs: 2)

    assert [(a, b) for a, b, func in path(foo.graph, A, C)] == [(A, B), (B, C)]

    ooc = set([A, C])
    assert [(a, b) for a, b, func in path(foo.graph, A, C, ooc_types=ooc)] == \
                        [(A, C)]
